# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Extension of :mod:`~.parse` module for debugging.

Main difference is the possibility of using the ``end_flag`` parameter in the
:mod:`~.scan`.
'''

import logging

from .parse import Parser, ParseResult, ParserException
from . import scan
from .grammar import t4debug_gram
from ...chrono import Chrono


LOGGER = logging.getLogger(__name__)


class ParserDebug(Parser):
    '''Scan up to the end flag then parse. For parsing debugging.'''

    def __init__(self, jddname, *, end_flag="", ofile=""):
        '''Initialize the :class:`ParserDebug` object.

        :param str jddname: path to the Tripoli-4 output
        :param int batch: batch to read (-1 = last, 0 = all, then X)
        :param str end_flag: optional end flag to stop scanning and parsing
                             (empty string per default)

        It also initalize the result of :class:`.scan.Scanner` to ``None`` and
        the parsing result, from *pyparsing*, to ``None``.

        If the path contains the ``"PARA"`` string, the checks will be done for
        parallel mode.
        '''
        self.end_flag = end_flag
        self.ofile = ofile
        super().__init__(jddname)

    def _scan_listing(self):
        '''Scan Tripoli-4 listing, calling :mod:`.scan`.'''
        return scan.Scanner(self.jdd, self.end_flag)

    def parse_from_number(self, batch_number, name=''):
        '''Parse from batch index or batch number.

        Per default the last batch is parsed (index = -1).

        :param int batch_number: the number of the batch to parse
        :returns: list(dict)
        '''
        LOGGER.debug('Using parse from ParserDebug')
        with Chrono() as chrono:
            batch_edition = self.scan_res[batch_number]
            if LOGGER.isEnabledFor(logging.DEBUG) and self.ofile:
                with open(self.ofile, 'w', encoding='utf-8') as fout:
                    fout.write(batch_edition)
            pres, = self._parse_listing_worker(
                t4debug_gram, self.scan_res[batch_number])
        LOGGER.info("Successful parsing in %f s", chrono)
        self.check_parsing(pres)
        try:
            self._time_consistency(pres, batch_number)
        except ParserException as tcve:
            if not self.end_flag:
                raise ParserException(tcve) from None
            LOGGER.info(tcve)
            LOGGER.info('Remark: you are in parsing debug mode with an end '
                        'flag not containing "time", this is expected.')
        scan_vars = self.scan_res.global_variables(batch_number)
        return ParseResult(pres, scan_vars, name=name)

    def check_parsing(self, parsed_res):
        '''Check if parsing went to the end:

        * if the end flag was not precised, a time should appear in the last
          result;
        * if not, no check can be performed as the end flag can be anywhere,
          even "transformed" during parsing.

        Print a logger message if not found but don't block access to the
        parsing result (this can help to find the issue).
        '''
        if not self.end_flag:
            if not any("_time" in s for s in parsed_res['batch_data']):
                LOGGER.error("Time not found in the parsing result, "
                             "parsing probably stopped before end, "
                             "please check.")
        if self.end_flag:
            LOGGER.info("You are running with an end flag ('%s'), "
                        "no automatic check of correct end of parsing is "
                        "currently available in that case, "
                        "please check carefully.", self.end_flag)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Common utilities for :program:`valjean` commands.'''

from pathlib import Path
import argparse
import sys
import inspect
import logging

from ..cosette.env import Env
from ..cosette.depgraph import DepGraph
from ..cosette.task import close_dependency_graph


LOGGER = logging.getLogger(__name__)


class Command:
    '''Base class for all :program:`valjean` subcommands.'''

    ALIASES = ()


class DictKwargAction(argparse.Action):
    '''An :class:`argparse.Action` subclass that parses arguments as
    ``key=value`` pairs and stores the resulting associations in a
    dictionary.'''

    def __call__(self, parser, namespace, option, option_string=None):
        '''Add a key-value pair to the dictionary.'''
        kwargs = getattr(namespace, self.dest)
        try:
            key, value = option.split('=', maxsplit=1)
        except ValueError:
            raise ValueError(f'cannot parse -k argument {option!r} as a '
                             'NAME=VALUE pair') from None
        kwargs[key] = value


class JobCommand(Command):
    '''Base class for all :program:`valjean` subcommands that take a job file
    and job arguments.'''

    def register(self, parser):
        '''Add the `job_file` and `job_args` positional arguments to the
        parser.'''
        parser.add_argument('job_file', action='store', metavar='JOB_FILE',
                            help='path to the job file')
        parser.add_argument('job_args', metavar='JOB_ARG', nargs='*',
                            help='positional arguments that will be passed to '
                            'the job() function; multiple arguments may be '
                            'given')
        parser.add_argument('-k', '--job-kwarg', metavar='NAME=VALUE',
                            dest='job_kwargs', action=DictKwargAction,
                            default={}, help='keyword arguments that will '
                            'be passed to the job() function; may be '
                            'specified multiple times')


def run_job(job_file, job_args, job_kwargs):
    '''Run the `job()` function from the specified job file and return its
    result.

    :param str job_file: the name of the file containing the `job()` function.
    :param list(str) job_args: the list of arguments to be passed to the
        `job()` function.
    :param dict job_kwargs: a dictionary of keyword arguments for `job()`
    :returns: whatever `job()` returns; expected to be a list of
        :class:`~.Task` objects.
    :rtype: list(Task)
    '''
    from ..dyn_import import dyn_import
    LOGGER.debug('importing job-file: %s', job_file)
    try:
        module = dyn_import(job_file)
    except FileNotFoundError:
        LOGGER.fatal('Cannot find job file %s', job_file)
        sys.exit(1)

    try:
        tasks = module.job(*job_args, **job_kwargs)
    except TypeError as err:
        if str(err).startswith('job()'):
            signature = inspect.signature(module.job)
            msg = ['argument mismatch to job() function',
                   f'  signature:\n    job{signature}']
            docstr = inspect.getdoc(module.job)
            if docstr is not None:
                docstr = docstr.replace('\n', '\n    ')
                msg.append(f'  docstring:\n    {docstr}')
            err = TypeError('\n'.join(msg))
        raise err
    LOGGER.debug('job tasks: %s', tasks)
    return tasks


def check_unique_task_names(tasks):
    '''Check that the tasks have unique names.

    :param list tasks: A list of tasks.
    :raises ValueError: if two or more tasks have the same name.
    '''
    names = set()
    dups = set()
    for task in tasks:
        if task.name in names:
            dups.add(task.name)
        names.add(task.name)
    if dups:
        dups_str = '\n  '.join(dups)
        raise ValueError('Task names must be unique; the following task names '
                         f'appear more than once:\n  {dups_str}')


def collect_tasks(job_file, job_args, job_kwargs):
    '''Collect tasks from a job file, along with all their dependencies.

    :param str job_file: the name of the file containing the `job()` function.
    :param list(str) job_args: the list of arguments to be passed to the
        `job()` function.
    :param dict job_kwargs: a dictionary of keyword arguments for `job()`
    :returns: the collected tasks.
    :rtype: list(Task)
    '''
    # import the job file and run the job() function
    tasks = run_job(job_file, job_args, job_kwargs)
    # compute the transitive closure of the dependency graph for the tasks
    # returned by job()
    tasks = close_dependency_graph(tasks)
    LOGGER.debug('collected tasks: %s', tasks)
    check_unique_task_names(tasks)
    return tasks


def build_graphs(args):
    '''Build the dependency graphs according to the CLI parameters.'''

    tasks = collect_tasks(args.job_file, args.job_args, args.job_kwargs)
    LOGGER.debug('building graphs for tasks: %s', tasks)

    hard_graph = DepGraph()
    soft_graph = DepGraph()
    for task in tasks:
        hard_graph.add_node(task)
        soft_graph.add_node(task)
        for dep in task.depends_on:
            hard_graph.add_dependency(task, on=dep)
        for dep in task.soft_depends_on:
            soft_graph.add_dependency(task, on=dep)

    LOGGER.debug('resulting hard_graph: %s', hard_graph)
    LOGGER.debug('resulting soft_graph: %s', soft_graph)
    return hard_graph, soft_graph


def read_env(*, root, names, filename, fmt):
    '''Create an initial environment for the given task names, possibly merging
    a set of serialized environments.

    The environment will be created from the partial environments that were
    serialized for the given task names. Missing partial environments will be
    silently ignored.

    If `filename` is `None`, no de-serialization will take place and an empty
    environment will be returned.

    :param str root: path to the root directory containing all the
        environment files.
    :param list(str) names: the list of task names that will be deserialized.
    :param filename: Name of the file containing the serialized environment. If
        `None`, no de-serialzation will take place.
    :type filename: str or None
    :param str fmt: Environment serialization format (only ``'pickle'`` is
        supported at the moment).
    :returns: an environment.
    :rtype: Env
    '''
    env = Env()
    if filename is None:
        return env
    LOGGER.info('deserializing %s environment from %r files in %s',
                fmt, filename, root)
    for task_name in names:
        task_file = str(Path(root) / task_name / filename)
        persisted_env = Env.from_file(task_file, fmt=fmt)
        if persisted_env is not None:
            env.merge_done_tasks(persisted_env)
    LOGGER.info('%d environment files found and deserialized', len(env))
    LOGGER.debug('deserialized environment: %s', env)
    return env


def write_env(env, *, filename, fmt):
    '''Serialize the environment to files.

    The environment will be written to one file per task (i.e. one per
    environment key). The name of the environment file is given by the
    `filename` parameter, and the directory is the output directory
    (``'output_dir'`` key) of the task. If the task does not have an
    ``'output_dir'`` key, serialization for that task will be skipped.

    If `filename` is `None`, no serialization will take place at all.

    :param filename: Name of the file containing the serialized environment. If
        `None`, no serialzation will take place.
    :type filename: str or None
    :param str fmt: Environment serialization format (only ``'pickle'`` is
        supported at the moment).
    '''
    if env is None or filename is None:
        LOGGER.debug('skipping environment serialization')
        return
    LOGGER.info('serializing %s environment to %r files', fmt, filename)
    LOGGER.debug('environment to serialize: %s', env)
    written_files = []
    for task_name, subenv in env.items():
        if 'output_dir' not in subenv:
            LOGGER.debug("skipping serialization of task %s because it does "
                         "not have any 'output_dir' key", task_name)
            continue
        task_file = str(Path(subenv['output_dir']) / filename)
        env.to_file(task_file, task_name=task_name, fmt=fmt)
        written_files.append(task_file)
    LOGGER.info('%d environment files written', len(written_files))
    LOGGER.debug('list of written environment files: %s', written_files)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Utility functions to access the filesystem.'''
from pathlib import Path


def ensure(*paths, is_dir=False):
    '''Make sure that the given path exists.

    :param paths: One or more paths. Multiple arguments will be concatenated
        into a single path.
    :type paths: str or pathlib.Path
    :param bool is_dir: If `True`, the path will be constructed as a directory.
    '''
    path = Path(*(str(path) for path in paths))
    if not path.exists():
        if is_dir:
            path.mkdir(parents=True)
        else:
            if not path.parent.exists():
                path.parent.mkdir(parents=True)
            path.touch()
    return path


def sanitize_filename(name):
    '''Check that the `name` string can be used as a filename.

    :raises ValueError: if the string contains characters that are forbidden in
        a filename.
    :returns: `name` unchanged
    '''
    if '\0' in name:
        raise ValueError(r"NULL character ('\0') is not allowed in filename "
                         f"{name!r}")
    if '/' in name:
        raise ValueError(f"slash ('/') is not allowed in filename {name!r}")
    if name in ('.', '..'):
        raise ValueError(f'{name!r} is not a valid filename')
    return name



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
'''The :class:`DepGraph` class encapsulates a number of useful methods to deal
with interdependent items. It represents the workhorse for the scheduling
algorithms in the :mod:`~.scheduler` module.

A dependency graph is a directed acyclic graph which may be represented as
follows:

.. digraph:: depgraph
   :align: center

   "bacon" -> "spam";
   "sausage" -> "eggs", "spam";

The convention here is that if an edge goes from `A` to `B`, then `A` depends
on `B`.


Building
--------

You can create a :class:`DepGraph` in of two ways. Either you pass a dictionary
representing the dependencies between items to the
:meth:`~DepGraph.from_dependency_dictionary()` class method:

   >>> from valjean.cosette.depgraph import DepGraph
   >>> from pprint import pprint
   >>> deps = {'bacon': ['spam'], 'sausage': ['eggs', 'spam']}
   >>> pprint(deps)
   {'bacon': ['spam'], 'sausage': ['eggs', 'spam']}
   >>> g = DepGraph.from_dependency_dictionary(deps)

or you create an empty graph and add nodes and dependencies by hand:

   >>> h = DepGraph().add_dependency('bacon', on='spam') \\
   ...               .add_dependency('sausage', on='eggs') \\
   ...               .add_dependency('sausage', on='spam')
   >>> g == h
   True

In these examples, `sausage` depends on `eggs` and `spam`, and `bacon` depends
on `spam`; `spam` and `eggs` have no dependencies. Note that the dependency
dictionary may be seen as a sparse representation of the graph adjacency
matrix.

You can recover the dependency dictionary by passing the graph to
:class:`dict() <dict>`:

   >>> pprint(dict(g))  # doctest: +SKIP
   {'bacon': {'spam'}, 'eggs': set(), 'sausage': {'eggs', 'spam'}, \
'spam': set()}

If you print the dictionary, you will notice that `spam` and `eggs` have been
added as nodes with no dependencies:

   >>> print(g)
   [('bacon', ['spam']), ('eggs', []), ('sausage', ['eggs', 'spam']), \
('spam', [])]

What if a node has no dependencies and no other node depends on it? Just add it
to the dictionary with the empty list as a value:

   >>> free = DepGraph.from_dependency_dictionary({'kazantzakis': []})

You can also add it after creating the graph:

   >>> also_free = DepGraph().add_node('kazantzakis')
   >>> free == also_free
   True


Querying
--------

You can inspect the nodes of the graph:

   >>> sorted(g.nodes())
   ['bacon', 'eggs', 'sausage', 'spam']

or ask for the dependencies of a node:

   >>> sorted(g.dependencies('sausage'))
   ['eggs', 'spam']
   >>> sorted(g['sausage'])  # equivalent, shorter syntax
   ['eggs', 'spam']

or ask for the nodes that depend on another node (careful though, this
operation has `O(N)` time complexity, `N` being the number of nodes in the
graph):

   >>> sorted(g.dependees('spam'))
   ['bacon', 'sausage']

You can also iterate over graphs:

   >>> for k, vs in sorted(g):
   ...     for v in sorted(vs):
   ...         print(f"You can't have {k} without {v}!")
   You can't have bacon without spam!
   You can't have sausage without eggs!
   You can't have sausage without spam!

Finally, you can check if a graph is a subgraph of another one:

   >>> sub_g = DepGraph().add_dependency('bacon', on='spam') \\
   ...                   .add_dependency('sausage', on='eggs')
   >>> sub_g <= g
   True


Merging, sorting and other operations
-------------------------------------

Given two graphs, possibly sharing some nodes and edges, you can construct the
*union* `g12` as follows:

   >>> g1 = sub_g
   >>> g2 = DepGraph().add_dependency('sausage', on='spam')
   >>> g12 = g1 + g2
   >>> g12 == g
   True
   >>> _ = g2.merge(g1)  # in-place merge
   >>> g2 == g
   True
   >>> g2 += g1          # equivalent syntax
   >>> g2 == g
   True

It is also possible to compute the transitive reduction of the graph. Let `g`
be an acyclic graph. The transitive reduction `tr(g)` is the minimal (in the
number of edges), provably unique subgraph of `g` over the same vertices with
the following property: for each pair of nodes `A` and `B`, `A` is reachable
from `B` within `g` iff it is reachable within `tr(g)`.

   >>> g_redundant = DepGraph() \\
   ...     .add_dependency('eggs', on='bacon') \\
   ...     .add_dependency('bacon', on='spam') \\
   ...     .add_dependency('eggs', on='spam')  # this edge is redundant
   >>> g_tr = g_redundant.copy()
   >>> print(g_tr.transitive_reduction())
   [('bacon', ['spam']), ('eggs', ['bacon']), ('spam', [])]
   >>> 'spam' in g_tr.dependencies('eggs')
   False

You can also do a topological sort of the graph. The result is a list of the
graph nodes, with the property that each node is guaranteed to appear after all
the nodes it depends on. Note that in general there are several possible
topological sorts for a given graph.

   >>> g.topological_sort()  # doctest: +SKIP
   ['eggs', 'spam', 'bacon', 'sausage']


Grafting sub-graph nodes into the graph itself
----------------------------------------------

A nice feature of :class:`DepGraph` is that you can have graph nodes that are
themselves instances of :class:`DepGraph`! Consider the following graph:

.. digraph:: depgraph
   :align: center

   compound=true;
   "Spanish Inquisition" -> "surprise" [lhead=cluster1];
   subgraph cluster1 {
     label="chief weapons"
     "surprise" -> "fear";
     "surprise" -> "efficiency";
     "fear" -> "devotion";
     "efficiency" -> "devotion";
   }
   "devotion" -> "nice red uniforms" [ltail=cluster1];

Here is the code that generates it:

   >>> spanish_inq = 'Spanish Inquisition'
   >>> surprise, fear, efficiency = 'surprise', 'fear', 'efficiency'
   >>> devotion, uniforms = 'devotion', 'nice red uniforms'
   >>> chief_weapons = DepGraph.from_dependency_dictionary({
   ...    surprise: [fear, efficiency],
   ...    fear: [devotion],
   ...    efficiency: [devotion]
   ...    })
   >>> spanish = DepGraph() \\
   ...     .add_dependency(spanish_inq, on=chief_weapons) \\
   ...     .add_dependency(chief_weapons, on=uniforms)

Here `chief_weapons` is a :class:`DepGraph` itself, but it is also a node of
`spanish`. You can graft `chief_weapons` inside `spanish` like so:

   >>> spanish.graft(chief_weapons)
   DepGraph(...)

This yields the following graph

.. digraph:: depgraph
   :align: center

   compound=true;
   "Spanish Inquisition" -> "surprise";
   "surprise" -> "fear";
   "surprise" -> "efficiency";
   "fear" -> "devotion";
   "efficiency" -> "devotion";
   "devotion" -> "nice red uniforms";

as you can verify yourself:

   >>> full_graph = DepGraph.from_dependency_dictionary({
   ...    spanish_inq: [surprise],
   ...    surprise: [fear, efficiency],
   ...    fear: [devotion],
   ...    efficiency: [devotion],
   ...    devotion: [uniforms]
   ...    })
   >>> full_graph == spanish
   True

The nice thing about this feature is that it makes it easier (more modular) to
build complex graphs. Just build smaller subgraphs and assemble them together
as if they were nodes! The :meth:`.flatten` method will recursively graft all
graph nodes into the main graph.

(Fun fact: the :class:`DepGraph` type is a monad, with :meth:`.flatten`
playing the role of ``join``. If what I just wrote makes no sense to you, don't
worry.)


Caveats
-------

Some things you should be aware of when using :class:`DepGraph`:

* The type of the items in the graph is irrelevant, but if you want to use the
  :meth:`~.DepGraph.from_dependency_dictionary()` constructor they need to be
  stored in a dictionary, and therefore they must be *hashable*;
* You need to use a single-element list if you want to express a single
  dependency, as in the case of `bacon`. So this is wrong:

    >>> bad_deps = {0: 1, 7: [0, 42]}  # error, should be 0: [1]
    >>> bad = DepGraph.from_dependency_dictionary(bad_deps)
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
        [...]
    TypeError: 'int' object is not iterable

  Here ``1`` is not iterable, and the test crashed. However, if your graph
  nodes happen to be iterable, the :class:`DepGraph` constructor will not
  crash, but you will not get what you expect.
'''

import logging
import itertools
import copy
import enum
from .rlist import RList
from ..chrono import Chrono


LOGGER = logging.getLogger(__name__)


class DepGraphError(Exception):
    '''An exception raised by :class:`DepGraph`.'''


class DepGraph:
    '''A dependency graph.

    There are two preferred ways to instantiate this class:

    * using the :meth:`from_dependency_dictionary()` class method;
    * constructing an empty graph and repeatedly calling
      :meth:`add_dependency()` and/or :meth:`add_node()`.

    Alternatively, you may also use the full form of the constructor; in this
    case, the graph `edges` are represented as a dictionary between integers,
    with the convention that each node is represented by its index in the
    `nodes` list.  So `nodes` can be seen as a mapping from integers to nodes.
    The inverse mapping is called `index` and may be passed to the constructor
    if it is available to the caller as a dictionary; if not, it will be
    constructed internally.
    '''
    _Marks = enum.Enum('Marks', 'TEMP PERM')

    @classmethod
    def from_dependency_dictionary(cls, dependencies):
        '''Generate a :class:`DepGraph` from a dependency dictionary.'''

        # the list of the graph nodes
        nodes = list(set(dependencies.keys())
                     | set(v for vs in dependencies.values() for v in vs))

        # translate the dependency dictionary elements to indices
        edges = {}
        for key, values in dependencies.items():
            new_key = nodes.index(key)
            new_values = set(map(nodes.index, values))
            edges[new_key] = new_values

        return cls(nodes, edges)

    @staticmethod
    def _complete(dct):
        '''Add singleton nodes to the dictionary.'''
        complete_dct = copy.copy(dct)
        for vals in dct.values():
            for val in vals:
                if val not in complete_dct:
                    complete_dct[val] = set()
        complete_dct = {k: set(v) for k, v in complete_dct.items()}
        return complete_dct

    def __init__(self, nodes=None, edges=None):
        '''Initialize the object from a list of nodes and an edge dictionary.

        :param nodes: An iterable over the nodes of the graph, or `None` for an
                      empty graph.
        :type nodes: :term:`iterable`
        :param edges: A mapping between integers representing the nodes, or
                      `None` for an empty graph.
        :type edges: :term:`mapping`
        '''

        if nodes is None or edges is None:
            self._nodes = RList()
            self._edges = {}
            return

        # self._nodes is the list of the graph nodes
        self._nodes = RList(nodes)
        LOGGER.debug('nodes: %s', self._nodes)

        # finally, complete the edges dictionary so that all values also appear
        # as keys, possibly with empty values
        LOGGER.debug('incomplete graph edges: %s', edges)
        with Chrono() as chrono:
            self._edges = DepGraph._complete(edges)
        LOGGER.debug('graph completed in %s seconds', chrono)
        LOGGER.debug('full graph edges: %s', self._edges)

    def __str__(self):
        assoc_list = sorted([(str(self._nodes[k]),
                              sorted(map(lambda v: str(self._nodes[v]), vs)))
                             for k, vs in self._edges.items()])
        return str(assoc_list)

    def __repr__(self):
        return f'DepGraph(nodes={self._nodes!r}, edges={self._edges!r})'

    def __iter__(self):
        for i, node in enumerate(self._nodes):
            values = map(lambda j: self._nodes[j], self._edges[i])
            yield (node, list(values))

    def __add__(self, other):
        '''Merge two graphs, return the result as a new graph.'''
        return self.copy().merge(other)

    __radd__ = __add__

    def __len__(self):
        '''Return the number of vertices in the graph.'''
        return len(self._nodes)

    def __contains__(self, node):
        '''Returns `True` if `x` is one of the nodes.'''
        return node in self._nodes

    def __le__(self, other):
        '''`g` <= `h` if `g` is a subgraph of `h`'''
        if not isinstance(other, DepGraph):
            return False
        # pylint: disable=protected-access
        return (all(node in other._nodes for node in self._nodes)
                and all(dep in other[node] for node in self._nodes
                        for dep in self[node]))

    def isomorphic_to(self, other):
        '''Returns `True` if this graph is isomorphic to `other`.'''

        if not isinstance(other, DepGraph):
            return False

        size = len(self)
        if size != len(other):
            return False
        i2o = [None] * size
        o2i = [None] * size
        for i, node in enumerate(self._nodes):
            # pylint: disable=W0212
            other_i = other._nodes.get_index(node, None)
            if other_i is None:
                LOGGER.debug('could not find node %s, not isomorphic', node)
                return False
            LOGGER.debug('node %s found: %d -> %d', node, i, other_i)
            i2o[i] = other_i
            o2i[other_i] = i
        if any(i is None for i in o2i):
            LOGGER.debug('some node is in other but not in self, '
                         'not isomorphic')
            return False
        for key, vals in self._edges.items():
            # pylint: disable=protected-access
            other_vals = set(o2i[o] for o in other._edges[i2o[key]])
            if set(vals) != set(other_vals):
                LOGGER.debug('node %d fails isomorphism check (%s != %s)',
                             key, vals, other_vals)
                return False
        LOGGER.debug('graphs are isomorphic')
        return True

    __eq__ = isomorphic_to

    def add_node(self, node):
        '''Add a new node to the graph.

        :param node: The new node.
        '''

        if node in self:
            LOGGER.debug('Node %s already belongs to the graph', node)
            return self

        new_index = len(self._nodes)
        self._nodes.append(node)
        self._edges[new_index] = set()
        return self

    def remove_node(self, node):
        '''Remove a node from the graph.

        Any edges going in and out of the node will be removed, too.

        :param node: The node to be removed.
        '''

        i = self._nodes.get_index(node, None)
        if i is None:
            LOGGER.warning('Cannot remove node %s: not in graph', node)
            return self

        # first, we need to put the selected node at the end of the _nodes list
        last = len(self._nodes) - 1
        self._nodes.swap(i, last)
        # swap nodes in the _edges dictionary
        tmp = self._edges[i]
        self._edges[i] = self._edges[last]
        self._edges[last] = tmp

        def _swapper(k):
            if k == i:
                return last
            if k == last:
                return i
            return k
        for k, vals in self._edges.items():
            self._edges[k] = set(map(_swapper, vals))
        # now we can remove the edges
        # remove edges from last
        del self._edges[last]
        # remove edges into last
        for k, vals in self._edges.items():
            self._edges[k] = set(n for n in vals if n != last)

        # finally, we can remove the node itself
        del self._nodes[last]

        return self

    def add_dependency(self, node, on):
        '''Add a new dependency to the graph.

        :param node: The node for which the dependency is specified.
        :param on: The node `node` depends on.
        '''

        self.add_node(node)
        self.add_node(on)
        i_node = self._nodes.index(node)
        i_on = self._nodes.index(on)
        self._edges[i_node].add(i_on)
        return self

    def remove_dependency(self, node, on):
        '''Remove a dependency from the graph.

        :param node: The node for which the dependency is specified.
        :param on: The node `node` depends on.
        '''

        i_node = self._nodes.index(node)
        i_on = self._nodes.index(on)
        try:
            self._edges[i_node].remove(i_on)
        except KeyError as kerr:
            raise KeyError(f'trying to remove missing edge {node} -> '
                           f'{on}') from kerr
        return self

    def invert(self):
        '''Invert the graph.

        :returns: A new graph having the same nodes but all edges inverted.
        '''

        inv_edges = {}
        for key, vals in self._edges.items():
            inv_edges.setdefault(key, [])
            for val in vals:
                inv_edges.setdefault(val, []).append(key)

        inverted = DepGraph(self._nodes, inv_edges)
        LOGGER.debug('inverted graph: %s', inverted)
        return inverted

    def topological_sort(self):
        '''Perform a topological sort of the graph.

        :returns: The nodes of the graph, as a list, with the invariant that
                  each node appears in the list after all the nodes it depends
                  on.
        :raises DepGraphError: if the graph is cyclic.
                  '''

        result = []
        marks = {}

        def _visit(node):
            mark = marks.get(node, None)
            if mark == self._Marks.TEMP:
                raise DepGraphError('Dependency graph is cyclic!')
            if mark == self._Marks.PERM:
                return
            marks[node] = self._Marks.TEMP
            index = self._nodes.index(node)
            for target in self._edges.get(index, []):
                _visit(self._nodes[target])
            marks[node] = self._Marks.PERM
            result.append(node)

        for node in self._nodes:
            if node in marks:
                continue
            _visit(node)

        return result

    def copy(self):
        '''Return a copy of this graph. '''
        return DepGraph(self._nodes.copy(), self._edges.copy())

    def merge(self, other):
        '''Merge this graph in place with another one.

        :param DepGraph other: The graph to merge.
        '''

        for key, vals in other:
            self.add_node(key)
            for val in vals:
                self.add_dependency(key, on=val)
        return self

    __iadd__ = merge

    def nodes(self):
        '''Returns the graph nodes.'''
        return self._nodes

    def dependencies(self, node, recurse=False):
        '''Query the graph about the dependencies of `node`. If `recurse` is
        ``True``, also return indirect dependencies (the transitive closure of
        the specified node). With `recurse=False`, this operation is `O(1)`.

        :param bool recurse: If true, return indirect dependencies as well.
        :returns: A list containing the dependencies of `node`.
        '''
        queue = [self._nodes.index(node)]
        seen = set()
        deps = set()
        while queue:
            next_node = queue.pop()
            seen.add(next_node)
            result = self._edges[next_node]
            LOGGER.debug('dependencies: %s -> %s', next_node, result)
            deps.update(result)
            if not recurse:
                break
            queue.extend(n for n in result if n not in seen)
        return [self._nodes[i] for i in deps]

    __getitem__ = dependencies

    def dependees(self, node):
        '''Collect the nodes that depend on the given node. This operation is
        `O(N)`, where `N` is the number of nodes in the graph.

        :returns: A set containing the dependees of `node`.
        '''
        i = self._nodes.index(node)
        indices = set(n for n in range(len(self)) if i in self._edges[n])
        return list(self._nodes[i] for i in indices)

    def to_graphviz(self):
        '''Convert the graph to graphviz format.

        :returns: A string describing the file in graphviz format.
        '''

        def sanitized_name(node):
            return str(node).replace('"', "'").replace('\n', '-')

        gviz = [r'digraph {']
        for key, vals in self._edges.items():
            from_node = sanitized_name(self._nodes[key])
            gviz.append('"' + from_node + '";')
            for val in vals:
                to_node = sanitized_name(self._nodes[val])
                gviz.append('"' + from_node + '" -> "' + to_node + '"')
        gviz.append(r'}')
        return '\n'.join(gviz)

    def transitive_reduction(self):
        '''Perform a transitive reduction of the graph in place.'''

        def _visit(start_edges, current):
            current_edges = self._edges[current]
            LOGGER.debug('visiting: %s: %s, %s',
                         current, current_edges, start_edges)
            to_remove = set()
            for dest in current_edges:
                if dest in start_edges:
                    LOGGER.debug('  marking %s for removal', dest)
                    to_remove.add(dest)
                LOGGER.debug('  recursing into %s', dest)
                to_remove |= _visit(start_edges, dest)
            return to_remove

        LOGGER.debug(repr(self))
        for i_node in range(len(self)):
            to_remove = set()
            for j_node in self._edges[i_node]:
                to_remove |= _visit(self._edges[i_node], j_node)
            for j in to_remove:
                self._edges[i_node].remove(j)

        return self

    def transitive_closure(self):
        '''Perform a transitive closure of the graph in place.'''

        def _visit(start_edges, current):
            current_edges = self._edges[current]
            LOGGER.debug('visiting: %s: %s, %s',
                         current, current_edges, start_edges)
            to_add = set()
            for dest in current_edges:
                if dest not in start_edges:
                    LOGGER.debug('  marking %s for insertion', dest)
                    to_add.add(dest)
                LOGGER.debug('  recursing into %s', dest)
                to_add |= _visit(start_edges, dest)
            return to_add

        LOGGER.debug(repr(self))
        for i_node in range(len(self)):
            to_add = set()
            for j_node in self._edges[i_node]:
                to_add |= _visit(self._edges[i_node], j_node)
            for j in to_add:
                self._edges[i_node].add(j)

        return self

    def initial(self):
        '''Return the initial nodes of the graph.

        The initial nodes are the nodes that have no ingoing edge; i.e., no
        other node depends on them.

        :returns: The list of initial nodes.
        '''

        index_set = set(range(len(self)))
        itarget_set = set()
        for vals in self._edges.values():
            itarget_set |= set(vals)
        index_set.difference_update(itarget_set)
        return [self._nodes[i] for i in index_set]

    def terminal(self):
        '''Return the terminal nodes of the graph.

        The terminal nodes are the nodes that have no outgoing edge; i.e., they
        do not depend on any other node.

        :returns: The list of terminal nodes.
        '''

        return [n for n in self._nodes if len(self[n]) == 0]

    def graft(self, node):
        '''Graft the given node into the graph.

        :param DepGraph node: A DepGraph embedded as a graph node.
        '''

        # collect the incoming and outgoing edges
        deps = self[node]
        dependees = self.dependees(node)
        # remove the graph-node
        self.remove_node(node)
        # graft the graph-node by adding incoming edges into its initial nodes
        # and outgoing edges from its terminal nodes
        inits = node.initial()
        terms = node.terminal()
        LOGGER.debug('Initial nodes = %s', inits)
        LOGGER.debug('Terminal nodes = %s', terms)
        LOGGER.debug('Dependencies = %s', deps)
        LOGGER.debug('Dependees = %s', dependees)
        self += node  # merge
        LOGGER.debug('Merged graph = %r', self)
        for dep in deps:
            for term in terms:
                LOGGER.debug('Adding dependency of %s on %s',
                             term, dep)
                self.add_dependency(term, on=dep)
        for dependee in dependees:
            for init in inits:
                LOGGER.debug('Adding dependency of %s on %s',
                             dependee, init)
                self.add_dependency(dependee, on=init)
        LOGGER.debug('Grafted graph = %r', self)

        return self

    def flatten(self, recurse=True):
        '''Graft all DepGraph nodes into this graph.

        :param bool recurse: If true, recursively graft DepGraph nodes until
                             all nodes are flat.
        '''
        nodes = [n for n in self._nodes if isinstance(n, DepGraph)]
        while nodes:
            LOGGER.debug('flatten() will graft the following nodes: %s',
                         nodes)
            for node in nodes:
                self.graft(node)
            if not recurse:
                break
            nodes = [n for n in self._nodes if isinstance(n, DepGraph)]
        return self

    def depends(self, node1, node2, recurse=False):
        '''Return ``True`` if the node `node1` depends on `node2`.

        :param node1: The first node.
        :param node2: The second node.
        :param bool recurse: If true, look at indirect dependencies, too.
        :returns: ``True`` if `node1` directly (``recurse == False``) or
                  indirectly (``recurse == True``) depends on `node2`.
        '''
        ind1 = self._nodes.index(node1)
        ind2 = self._nodes.index(node2)
        deps1 = self._edges[ind1]
        depends = False
        while deps1:
            depends = ind2 in deps1
            if depends or not recurse:
                return depends
            deps1 = list(itertools.chain.from_iterable(self._edges[i]
                                                       for i in deps1))
        return depends



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module converts Tripoli-4 data from parsing in
:class:`~valjean.eponine.dataset.Dataset`.
'''

import logging
from collections import OrderedDict
import numpy as np
from ..dataset import Dataset


LOGGER = logging.getLogger(__name__)


def bins_reduction(def_bins, reduced_dims):
    '''Reduce number of bins for integrated results (on one or more dimension).

    For the reduced dimensions, all the bounds are suppressed except the first
    and the last one. This is also the case if they represent center of bins.
    The dimension is supposed to be squeezed afterwards.

    :param def_bins: default bins (all of them)
    :type def_bins: :obj:`collections.OrderedDict` (str, :obj:`numpy.ndarray`)
    :param list(bool): dimensions to be reduced or not
    :returns: :obj:`collections.OrderedDict` (str, :obj:`numpy.ndarray`), i.e.
        adapted bins

    >>> ibins = OrderedDict([('a', np.array([0, 1, 2, 3, 4])),
    ...                      ('b', np.array([-10, 0, 10]))])
    >>> dshape_1 = (5, 3)
    >>> b11 = bins_reduction(ibins, [True, True])
    >>> b11['a']
    array([0, 4])
    >>> b11['b']
    array([-10,  10])
    >>> b51 = bins_reduction(ibins, [False, True])
    >>> b51['a']
    array([0, 1, 2, 3, 4])
    >>> b51['b']
    array([-10,  10])
    '''
    bins = OrderedDict()
    for ibin, _bin in enumerate(def_bins.keys()):
        if reduced_dims[ibin]:
            bins[_bin] = def_bins[_bin][::def_bins[_bin].size-1]
        else:
            LOGGER.debug("Keep bins from 'array'.")
            bins[_bin] = def_bins[_bin]
    return bins


def special_array(array, score, bins, array_key='array', name='', what=''):
    '''Convert special arrays in :class:`~.dataset.Dataset`.

    These special cases are typically vov results or uncertainty spectrum
    associated to a perturbation result. No error is associated to the score in
    that case. Bins are given by the spectrum.

    :param dict array: result
    :param str score: key of the score to get
    :param collections.OrderedDict bins: bins correspondig to required array
    :param str array_key: default=``'array'`` but it can be an integrated
        array for example, should be a key inside ``farray_res``
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :rtype: Dataset
    '''
    the_array = array[array_key][score]
    return Dataset(
        the_array.copy(), np.full_like(the_array, np.nan), bins=bins,
        name=name, what=what)


def array_result(farray_res, res_type, name='', what='', array_key='array',
                 score='score'):
    '''Conversion of arrays in :class:`~.dataset.Dataset`.

    :param dict farray_res: results dictionary containing ``res_type`` key
    :param str res_type: result type, like ``'spectrum'``, ``'mesh'``
    :param str array_key: default=``'array'`` but it can be an integrated
        array for example, should be a key inside ``farray_res``
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :returns: :class:`~.dataset.Dataset`
    '''
    array_res = farray_res[res_type]
    if 'array' not in farray_res[res_type]:
        raise KeyError("key 'array' should be in the available keys")
    if array_key not in array_res:
        LOGGER.warning("array_key %s not found", array_key)
        return None
    bins = (array_res.get('bins') if array_key == 'array'
            else bins_reduction(array_res.get('bins'),
                                [o != d
                                 for d, o in zip(array_res['array'].shape,
                                                 array_res[array_key].shape)]))
    if score != 'score':
        return special_array(array_res, score, bins, array_key, name, what)
    return Dataset(
        array_res[array_key][score].copy(),
        array_res[array_key]['sigma'] * array_res[array_key][score] * 0.01,
        bins=bins, name=name, what=what)


def masked_array_result(farray_res, res_type, name='', what='',
                        array_key='array', score='score'):
    '''Mask invalid (``np.nan`` and ``np.inf``) values.'''
    ards = array_result(farray_res, res_type, name=name, what=what,
                        array_key=array_key, score=score)
    mask = np.ma.masked_invalid(ards.value).mask  # pylint: disable=no-member
    return ards.mask(mask)


def integrated_result(result, res_type='integrated', name='', what='',
                      score='score', sigma='sigma'):
    '''Conversion of generic score (or energy integrated result) in
    :class:`~valjean.eponine.dataset.Dataset`.

    Bins are squeezed according to the integrated dimension obtained from
    another array stored in the same result. Default full arrays are
    `'spectrum'` and `'mesh'`. A case is also foreseen for uncertainty
    results (in perturbation cases).

    :param dict result: results dictionary containing ``res_type`` key
    :param str res_type: should be 'integrated'
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :param str score: score key in results array like `'score'` or
        `'score/lethargy'`
    :param str sigma: sigma key in results array, usually `'sigma'`
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    LOGGER.debug("In integrated_result")
    intres = result[res_type] if res_type in result else result
    if isinstance(intres, dict) and 'not_converged' in intres:
        return nan_result(name=name, what=what)
    sstring = 'uncert' if 'uncert' in res_type else ''
    other_res = [x for x in result
                 if (x != res_type and ('spectrum' in x or 'mesh' in x)
                     and sstring in x)]
    full_res = other_res[0]
    if len(other_res) > 1:
        LOGGER.warning('More than one result can have been integrated %s, '
                       'chosen one: %r for %r', other_res, full_res, res_type)
    ishape = tuple([1]*result[full_res]['array'].ndim)
    diffshape = [o != d for o, d in zip(result[full_res]['array'].shape,
                                        ishape)]
    bins = bins_reduction(result[full_res]['bins'], diffshape)
    return Dataset(np.full(ishape, intres[score]),
                   (np.full(ishape, intres[sigma]*intres[score]*0.01) if sigma
                    else np.full(ishape, np.float_(np.nan))),
                   bins=bins, name=name, what=what)


def results_wo_error(result, res_type, name='', what='', score=None):
    '''Conversion of results without error, allowing multiple scores in the
    same container.

    :dict result: results dictionary containing ``res_type`` key
    :param str res_type: key in result allowing access to the desired quantity
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :param str score: score key in results array like `'score'`
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    if score:
        return result_wo_error(result[res_type], score, name=name, what=what)
    return result_wo_error(result, res_type, name=name, what=what)


def result_wo_error(result, res_type, name='', what=''):
    '''Conversion of a result without error.

    The result can be of any dimension (matrix, vector, scalar) or any type
    (float, int, complex).

    :dict result: results dictionary containing ``res_type`` key
    :param str res_type: key in result allowing access to the desired quantity
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    LOGGER.debug("value without error %s", result[res_type])
    res = result[res_type]
    error = 0 if np.issubdtype(res.dtype, np.integer) else np.nan
    return Dataset(res, np.full(res.shape, error), name=name, what=what)


def result_with_error(result, res_type, name='', what='',
                      score='score', sigma='sigma'):
    '''Conversion of generic score (or energy integrated result) in
    :class:`~valjean.eponine.dataset.Dataset`.

    :param dict result: results dictionary containing ``res_type`` key
    :param str res_type: should be 'generic'
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :param str score: score key in results array like `'score'` or
        `'score/lethargy'`
    :param str sigma: sigma key in results array, usually `'sigma'`
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    LOGGER.debug("result_with_error")
    intres = result[res_type]
    if isinstance(intres, dict) and 'not_converged' in intres:
        return nan_result(name=name, what=what)
    if set(('sigma', 'sigma%')).issubset(intres):
        # used for mean weight leakage or automatic keff for example
        return Dataset(intres[score].copy(), intres['sigma'].copy(),
                       name=name, what=what)
    return Dataset(intres[score].copy(),
                   intres[sigma] * intres[score] * 0.01,
                   name=name, what=what)


def unbinned_result(result, res_type, name='', what='', score=None,
                    sigma=None):
    '''Conversion of all unbinned results in
    :class:`~valjean.eponine.dataset.Dataset`.

    :dict result: results dictionary containing ``res_type`` key
    :param str res_type: key in result allowing access to the desired quantity
    :param str name: name of dataset
    :param str what: what attribute of dataset
    :param str score: score key in results array like `'score'` or
        `'score/lethargy'`
    :param str sigma: sigma key in results array, usually `'sigma'`
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    if sigma:
        return result_with_error(result, res_type, name=name, what=what,
                                 score=score, sigma=sigma)
    return results_wo_error(result, res_type, name=name, what=what,
                            score=score)


def nan_result(name='', what=''):
    '''Returns a NaN :class:`~.dataset.Dataset` (value and error).

    This :class:`~.dataset.Dataset` can be returned for example in
    case of non converged results or when no dataset can be built.

    Values are scalar per default, not arrays.

    :param str name: name of dataset
    :param str what: what attribute of dataset
    :rtype: :class:`~.dataset.Dataset`
    :returns: dataset containing NaN as value and error
    '''
    LOGGER.info('Result not converged, return dataset with NaN value')
    return Dataset(np.nan, np.nan, name=name, what=what)


CONVERT_IN_DATASET = {
    'spectrum': array_result,
    'uncert_spectrum': array_result,
    'uncert_integrated': integrated_result,
    'mesh': array_result,
    'green_bands': array_result,
    'sensitivity_spectrum': array_result,
    'adj_crit_ed': array_result,
    'integrated': integrated_result,
    'best_result': integrated_result,
    'spherical_harmonics': masked_array_result
}


def convert_data(data, data_type, name='', what='', **kwargs):
    '''Test for data conversion using dict or default.

    An exception for integrated is for the moment needed as they can come
    from spectrum res or generic scores but are treated a bit differently.
    To be homogenized.

    :param dict data: results dictionary containing ``data_type`` key
    :param str data_type: data key
    :param kwargs: keyword arguements if needed
    :returns: :class:`~valjean.eponine.dataset.Dataset`
    '''
    if data_type not in data:
        if 'not_converged' in data:
            return nan_result(name, what)
        LOGGER.warning("%s not found in data", data_type)
        return None
    if isinstance(data[data_type], str):
        if 'not_converged' in data[data_type] or data_type == 'not_converged':
            return nan_result(name, what)
        LOGGER.warning("Dataset cannot be built from a string.")
        return nan_result(name, what)
    return CONVERT_IN_DATASET.get(data_type, unbinned_result)(
        data, data_type, name, what, **kwargs)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''This module provides a decorator that simplifies the integration of free
Python functions in the :mod:`valjean` dependency graph.


An example
==========

Let us consider the following simple function, that counts the number of
characters in a file:

    >>> def how_many_chars(filename):
    ...     return len(open(filename).read())

Suppose that we have a task that produces some text and we want to count the
number of characters in the output. For this example, we will take the task to
be:

    >>> from valjean.cosette.run import RunTask
    >>> cmd = ['echo', 'spam']
    >>> run_task = RunTask.from_cli('run_task', cmd)

This task will produce exactly five characters of standard output [#]_, but
assume we want to check this with our `how_many_chars` function. If the task is
part of a dependency graph, we cannot call `how_many_chars` until the task has
been executed, simply because its output (which is the input to
`how_many_chars`) does not exist until then.

The solution is to bridge the gap using an :class:`Use` object:

    >>> from valjean.cosette.use import Use
    >>> use_arg_echo = Use.from_func(func=how_many_chars, task=run_task)

:class:`Use` lifts the naked function `how_many_chars` into the world of
:mod:`valjean` dependency graphs. The `how_many_chars` function is wrapped in a
:class:`~.PythonTask` that can be retrieved with the :meth:`~.get_task`
method:

    >>> how_many_task = use_arg_echo.get_task()
    >>> from valjean.cosette.pythontask import PythonTask
    >>> isinstance(how_many_task, PythonTask)
    True

The `how_many_task` object is wired in such a way that it will retrieve the
value to its argument from the ``'stdout'`` key of `run_task`'s environment
section on execution. If `run_task` has not been executed at that point or it
has failed, then of course `how_many_task` will fail, too.  The dependency of
`how_many_task` on `run_task` is made explicit:

    >>> run_task in how_many_task.depends_on
    True

.. note::

    If you want to introduce a **soft** dependency on `run_task` (instead of a
    hard one), you can use the `deps_type` constructor argument:

    >>> use_arg_echo_soft = Use.from_func(func=how_many_chars, task=run_task,
    ...                                   deps_type='soft')
    >>> how_many_task_soft = use_arg_echo_soft.get_task()
    >>> run_task in how_many_task_soft.soft_depends_on
    True

This allows :mod:`valjean` to seamlessly integrate the execution of
`how_many_task` in the dependency graph.  We can illustrate the relations
between the objects that we have created with the following diagram:

.. digraph:: depgraph
    :align: center

    compound=true;
    subgraph cluster_use {
        label="use_arg_echo: Use";
        func [label="how_many_chars: func" shape=diamond];
    }
    run_task [label="run_task: Task"];
    how_many_task [label="how_many_task: Task"]
    { rank=same run_task how_many_task }
    func -> how_many_task [label=" generates" style=dotted ltail=cluster_use];
    how_many_task -> run_task [label="depends on"];

Here we see that `use_arg_echo` has type :class:`Use` and contains the
`how_many_chars` function. Additionally, it generates `how_many_task`, which is
a :class:`~.Task` object that depends on `run_task`.


.. doctest:: cosette_use
    :hide:

    >>> from valjean.config import Config
    >>> config = Config()

We can test that `how_many_task` works by manually executing `run_task` first
and passing the resulting environment to `how_many_task`:

    >>> from valjean.cosette.env import Env
    >>> env, status = run_task.do(env=Env(), config=config)
    >>> print(status)
    TaskStatus.DONE

Note that the environment now contains the name of the file written by
`run_task` in the ``'stdout'`` key:

    >>> env[run_task.name]['stdout']
    '/.../run_task/stdout'

We can now execute `how_many_task`:

    >>> env_up, status = how_many_task.do(env=env, config=config)
    >>> print(status)
    TaskStatus.DONE

The ``'result'`` key now holds the return value of `how_many_chars`:

    >>> env_up[how_many_task.name]['result']
    5


:class:`Use` objects are wrappers
---------------------------------

:class:`Use` also behaves as a wrapper of `how_many_chars`.  You can still call
the underlying wrapped function with explicit parameters if you wish:

    >>> with open('test.txt', 'w') as f:
    ...     _ = f.write('lobster Thermidor\n')
    >>> use_arg_echo('test.txt')
    18
    >>> len('lobster Thermidor\n')
    18

This simplifies testing.


Argument injection via the :func:`using` decorator
--------------------------------------------------

A practical way to create :class:`Use` objects is to call the :func:`using`
decorator:

    >>> @using(task=run_task)
    ... def how_many_chars(filename):
    ...     return len(open(filename).read())

Now `how_many_chars` is itself a :class:`Use` object:

    >>> isinstance(how_many_chars, Use)
    True

This is exactly equivalent to what we did earlier, except for the fact that the
:class:`Use` object now does not have a different name:

.. digraph:: depgraph
    :align: center

    compound=true;
    subgraph cluster_use {
        label="how_many_chars: Use";
        fontcolor=red;
        func [label="how_many_chars: func" shape=diamond];
    }
    run_task [label="run_task: Task"];
    how_many_task [label="how_many_task: Task"]
    { rank=same run_task how_many_task }
    func -> how_many_task [label=" generates" style=dotted ltail=cluster_use];
    how_many_task -> run_task [label="depends on"];

You can still call the underlying `how_many_chars` by calling the :class:`Use`
object as if it were a normal function, as discussed above. Just like before,
we can generate a task from the :class:`Use` object:

    >>> how_many_task = how_many_chars.get_task()
    >>> isinstance(how_many_task, PythonTask)
    True


Injecting multiple arguments
----------------------------

If your function expects multiple arguments to be fed from previous tasks,
you can stack multiple :class:`Use` decorators:

    >>> name_task = RunTask.from_cli('name_task', ['echo', 'Arthur'])
    >>> job_task = RunTask.from_cli('job_task',
    ...                             ['echo', 'King of the Britons'])
    >>> @using(task=name_task)
    ... @using(task=job_task)
    ... def introduce(name_filename, job_filename):
    ...     with open(name_filename) as name_file:
    ...         name = name_file.read().rstrip()
    ...     with open(job_filename) as job_file:
    ...         job = job_file.read().rstrip()
    ...     return 'I am ' + name + ', ' + job + '!'

The resulting task has the correct dependencies, i.e. it depends on both
`name_task` and `job_task`:

    >>> intro_task = introduce.get_task()
    >>> name_task in intro_task.depends_on
    True
    >>> job_task in intro_task.depends_on
    True

We can represent the dependency structure of the generated tasks with the
following graph:

.. digraph:: depgraph
    :align: center

    compound=true;
    subgraph cluster_intro {
      label="introduce: Use";
      introduce [label="introduce: func" shape=diamond];
    }
    name [label="name_task: Task"];
    job [label="job_task: Task"];
    "intro_task: Task" -> name [label="depends on"];
    "intro_task: Task" -> job [label="depends on"];
    introduce -> "intro_task: Task"
      [label=" generates" style=dotted ltail=cluster_intro];


If we manually execute `name_task` and `job_task`, we can then execute
`intro_task`:

    >>> env = Env()
    >>> for task in [name_task, job_task, intro_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[intro_task.name]['result'])
    I am Arthur, King of the Britons!

Keyword arguments
-----------------

In the previous example, note that the order of the decorators is important!
The outer decorator (`name_task`) injects the first argument (`name_filename`),
and the inner one (`job_task`) injects the second argument (`job_filename`). If
we nest the decorators the other way around, we inverse the order of the
arguments:

    >>> @using(task=job_task)
    ... @using(task=name_task)
    ... def introduce_inv(name_filename, job_filename):
    ...     with open(name_filename) as name_file:
    ...         name = name_file.read().rstrip()
    ...     with open(job_filename) as job_file:
    ...         job = job_file.read().rstrip()
    ...     return 'I am ' + name + ', ' + job + '!'
    >>> intro_inv_task = introduce_inv.get_task()
    >>> for task in [name_task, job_task, intro_inv_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[intro_inv_task.name]['result'])
    I am King of the Britons, Arthur!

If you are injecting many arguments, it may be a good idea to use the `kwarg`
argument to :func:`using`, which allows you to inject values into arguments by
name instead of by position:

    >>> @using(kwarg='name_filename', task=name_task)
    ... @using(kwarg='job_filename', task=job_task)
    ... def introduce_kwargs(name_filename, job_filename):
    ...     return introduce(name_filename, job_filename)
    >>> @using(kwarg='job_filename', task=job_task)
    ... @using(kwarg='name_filename', task=name_task)
    ... def introduce_kwargs_swap(name_filename, job_filename):
    ...     return introduce(name_filename, job_filename)

Note that the `introduce_kwargs` and `introduce_kwargs_swap` functions are
identical, but the `@using` decorators appear in reverse order. In the previous
example, this mattered and we ended up with ``I am King of the Britons,
Arthur!``. But since now we are using the `kwarg` argument, the order of the
decorators does not matter any more.  Both `introduce_kwargs` and
`introduce_kwargs_swap` will inject the arguments as expected:

    >>> intro_kwargs_task = introduce_kwargs.get_task()
    >>> intro_kwargs_swap_task = introduce_kwargs_swap.get_task()
    >>> env_up, _ = intro_kwargs_task.do(env=env, config=config)
    >>> print(env_up[intro_kwargs_task.name]['result'])
    I am Arthur, King of the Britons!
    >>> env_up, _ = intro_kwargs_swap_task.do(env=env, config=config)
    >>> print(env_up[intro_kwargs_swap_task.name]['result'])
    I am Arthur, King of the Britons!


Modifying an existing :class:`Use` decorator
--------------------------------------------

You can pipe the result of the task generated by a :class:`Use` decorator by
using the :meth:`Use.map` method:

    >>> intro_stars = introduce.map(lambda x: '★★★' + x + '★★★')

Here `intro_stars` is a new :class:`Use` object that will apply the specified
function (here, a lambda that surrounds its argument with ``★★★`` cute little
stars ``★★★``) to the result of `intro_task` (see above). The task associated
to `intro_stars` depends on `intro_task`:

    >>> intro_stars_task = intro_stars.get_task()
    >>> intro_task in intro_stars_task.depends_on
    True

In our schematic representation, `intro_stars` and `intro_stars_task` would
appear like this:

.. digraph:: depgraph
    :align: center

    compound=true;
    subgraph cluster_intro {
        label="introduce: Use";
        introduce [label="introduce: func" shape=diamond];
    }
    name [label="name_task: Task"];
    job [label="job_task: Task"];
    intro_task [label="intro_task: Task"];
    intro_task -> name [label="depends on"];
    intro_task -> job [label="depends on"];
    introduce -> intro_task
      [label=" generates" style=dotted ltail=cluster_intro];

    subgraph cluster_stars {
        label="intro_stars: Use";
        color=red;
        fontcolor=red;
        stars [label=": <lambda>" shape=diamond color=red fontcolor=red];
    }
    intro_stars_task
      [label="intro_stars_task: Task" color=red fontcolor=red];
    stars -> intro_stars_task
      [label=" generates" style=dotted ltail=cluster_stars];
    intro_stars_task -> intro_task [label="depends on"];

    { rank=same intro_task intro_stars_task }


If we execute `intro_stars_task`, we find that our string now has stars!

    >>> env_up, _ = intro_stars_task.do(env=env, config=config)
    >>> env.apply(env_up)
    >>> print(env[intro_stars_task.name]['result'])
    ★★★I am Arthur, King of the Britons!★★★


.. rubric:: Footnotes

.. [#] The proof is left to the reader as an exercise.


Injecting arguments into tasks created by a :class:`~.RunTaskFactory`
=====================================================================

There is a typical use case for :class:`Use` which consists in injecting the
result of a :class:`~.RunTask` generated by a :class:`~.RunTaskFactory` as an
argument to a subsequent task `B`. Since this use case is common, there is a
special :class:`UseRun` class that helps reducing the boilerplate.

First, we create a :class:`~.RunTaskFactory` object:

    >>> from valjean.cosette.run import RunTaskFactory
    >>> factory = RunTaskFactory.from_executable('echo')

Instead of creating :class:`~.RunTask` objects and wrapping them in
:class:`Use` by hand, we instantiate a :class:`UseRun` object:

    >>> using_run = UseRun.from_factory(factory)

The :class:`UseRun` object will create the :class:`~.RunTask` objects on the
fly, when invoked as a decorator:

    >>> @using_run(extra_args=['spam'])
    ... def read_and_check_text(filename):
    ...     with open(filename) as f:
    ...         return f.read() == 'spam\n'

The decorated function is simply an instance of the :class:`Use` class that we
all know and love:

    >>> isinstance(read_and_check_text, Use)
    True

so we can call :meth:`Use.get_task` to generate a task:

    >>> read_and_check_task = read_and_check_text.get_task()
    >>> run_task = next(iter(read_and_check_task.depends_on))

Since a picture is worth a thousand words, the objects have the following
structure:

.. digraph:: depgraph
    :align: center

    compound=true;
    using_run [label="using_run: UseRun" shape=box];
    subgraph cluster_read {
        label="read_and_check_text: Use";
        read [label="read_and_check_text: func" shape=diamond];
        factory [label="factory: RunTaskFactory"];
    }
    using_run -> read [label=" generates" lhead=cluster_read style=dotted];
    factory -> run_task [label="generates" style=dotted];
    read_task [label="read_and_check_task: Task"];
    run_task [label="run_task: Task"];
    read_task -> run_task [label="depends on"];
    read -> read_task [label=" generates" style=dotted ltail=cluster_read];
    { rank=same read_task run_task }

Let's check that our tasks work as expected:

    >>> env, _ = run_task.do(env=Env(), config=config)
    >>> env, _ = read_and_check_task.do(env=env, config=config)
    >>> print(env[read_and_check_task.name]['result'])
    True


Pipelines
---------

This is already nice! However, if you often want to perform the same additional
actions after creating your :class:`~.RunTask`, you can cut the boilerplate
further with the :meth:`UseRun.map` method. In our example, we might want to
factorize out the ``with open(filename) as f`` bit and focus in the text
comparison:

    >>> def slurp(filename):
    ...     with open(filename) as f:
    ...         return f.read()
    >>> using_run_as_text = using_run.map(slurp)

The  :meth:`UseRun.map` method creates a new :class:`UseRun` object which will
enqueue the specified function (as a task) to be executed on the result of the
:class:`~.RunTask`. We can simplify `read_and_check_text` from above as
follows:

    >>> @using_run_as_text(extra_args=['spam'])
    ... def check_text(text):
    ...     return text == 'spam\n'

The diagram now looks like this:

.. digraph:: depgraph
    :align: center

    compound=true;
    using_run [label="using_run: UseRun" shape=box];
    using_run_as_text [label="using_run_as_text: UseRun" shape=box];
    using_run -> using_run_as_text [label=" map(slurp)" style=dotted];
    subgraph cluster_check {
        label="check_text: Use";
        slurp [label="slurp: func" shape=diamond];
        check [label="check_text: func" shape=diamond];
        factory [label="factory: RunTaskFactory"];
        { rank=same slurp check }
    }
    using_run_as_text -> slurp
      [label=" generates" lhead=cluster_check style=dotted];
    factory -> run_task [label=" generates" style=dotted];
    check_task [label="check_task: Task"];
    slurp_task [label="slurp_task: Task"];
    run_task [label="run_task: Task"];
    check_task -> slurp_task [label="depends on"];
    slurp_task -> run_task [label="depends on"];
    slurp -> slurp_task [label=" generates" style=dotted ltail=cluster_check];
    check -> check_task [label=" generates" style=dotted ltail=cluster_check];
    { rank=same slurp_task check_task run_task }

Checking that this works is a bit more laborious, because we now have one extra
task to run; in the real world, of course, :mod:`valjean` would take care of
running the tasks via a dependency graph and we wouldn't need to worry about
any of this:

    >>> check_task = check_text.get_task()
    >>> slurp_task = next(iter(check_task.depends_on))
    >>> run_task = next(iter(slurp_task.depends_on))
    >>> env = Env()
    >>> for task in [run_task, slurp_task, check_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[check_task.name]['result'])
    True

You can also chain :meth:`UseRun.map` calls:

    >>> using_run_as_stripped_text = using_run.map(slurp) \
    ...                                       .map(lambda x: x.strip())
    >>> @using_run_as_stripped_text(extra_args=['spam'])
    ... def check_stripped_text(text):
    ...     return text == 'spam'  # note that the \n is gone

Proof that it works:

    >>> check_stripped_task = check_stripped_text.get_task()
    >>> strip_task = next(iter(check_stripped_task.depends_on))
    >>> slurp_task = next(iter(strip_task.depends_on))
    >>> run_task = next(iter(slurp_task.depends_on))
    >>> env = Env()
    >>> for task in [run_task, slurp_task, strip_task, check_stripped_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[check_stripped_task.name]['result'])
    True

and a final diagram to show the dependencies among the objects:

.. digraph:: depgraph
    :align: center

    compound=true;
    using_run [label="using_run: UseRun" shape=box];
    using_run_anonymous
      [label="<anonymous>: UseRun" shape=box];
    using_run_as_stripped_text
      [label="using_run_as_stripped_text: UseRun" shape=box];
    using_run -> using_run_anonymous [label=" map(slurp)" style=dotted];
    using_run_anonymous -> using_run_as_stripped_text
      [label=" map(strip)" style=dotted];
    subgraph cluster_check_stripped {
        label="check_stripped_text: Use";
        slurp [label="slurp: func" shape=diamond];
        strip [label="strip: func" shape=diamond];
        check_stripped [label="check_stripped_text: func" shape=diamond];
        factory [label="factory: RunTaskFactory"];
        { rank=same slurp check_stripped strip }
    }
    using_run_as_stripped_text -> slurp
      [label=" generates" lhead=cluster_check_stripped style=dotted];
    factory -> run_task [label=" generates" style=dotted];
    check_stripped_task [label="check_stripped_task: Task"];
    strip_task [label="strip_task: Task"];
    slurp_task [label="slurp_task: Task"];
    run_task [label="run_task: Task"];
    check_stripped_task -> strip_task [label="depends on"];
    strip_task -> slurp_task [label="depends on"];
    slurp_task -> run_task [label="depends on"];
    strip -> strip_task
      [label=" generates" style=dotted ltail=cluster_check_stripped];
    slurp -> slurp_task
      [label=" generates" style=dotted ltail=cluster_check_stripped];
    check_stripped -> check_stripped_task
      [label=" generates" style=dotted ltail=cluster_check_stripped];
    { rank=same slurp_task check_stripped_task run_task strip_task }


Module API
==========
'''

import logging
from pathlib import Path
from functools import update_wrapper

from .task import TaskStatus
from ..path import ensure, sanitize_filename
from .pythontask import PythonTask


LOGGER = logging.getLogger(__name__)


def from_env(*, env, task_name, key):
    '''Helper function to extract a value from the environment for argument
    injection.

    :param env: the environment.
    :type env: :class:`~.Env`
    :param str task_name: the name of the task to look up.
    :param key: the key to look up. If `None` is given, return the whole
        key-value pair associated to `task_name`.
    :type key: str or None
    :raises KeyError: if the required task_name or key are not available.
    '''
    try:
        task_env = env[task_name]
    except KeyError:
        LOGGER.error('Task %r is required for argument injection, but the '
                     'task is not in the environment', task_name)
        raise

    if key is None:
        return task_name, task_env

    try:
        task_result = task_env[key]
    except KeyError:
        LOGGER.error('Task %r is required for argument injection, but I '
                     'could not find the %r key in the result environment',
                     task_name, key)
        raise
    return task_result


class Use:
    '''A function wrapper around a free Python function. Lifts the function
    into a :mod:`~.valjean` :class:`~.Task`.'''

    _CACHE = {}
    _USE_CACHING = True

    @classmethod
    def from_func(cls, *, func, task, key='result', kwarg=None,
                  deps_type='hard', serialize=False):
        # pylint: disable=too-many-arguments
        '''Create a :class:`Use` from a function.

        :param func: a function or a callable object.
        :param task: the task whose result should be injected as an argument to
            `func`.
        :type task: :class:`~.Task`
        :param str key: the name of the key that contains the task result in
            the environment.
        :param kwarg: the name of the keyword argument to `func` that must be
            fed with the `task` result. If None, the result of `task` will be
            passed as a positional argument.
        :type kwarg: None or str
        :param str deps_type: whether the created task should have a hard or a
            soft dependency towards the injected tasks. Possible values are
            ``'hard'`` and ``'soft'``.
        :param bool serialize: if `True`, the result of this task will be
            written to the output directory.
        '''
        if isinstance(func, cls):
            # when we decorate another Use object, we just extend the decorated
            # object's inj_args list and inj_kwargs dictionary
            LOGGER.debug('decorating a "%s" object', cls.__name__)
            inj_args = func.inj_args.copy()
            inj_kwargs = func.inj_kwargs.copy()
        else:
            LOGGER.debug('decorating a free function')
            inj_args = []
            inj_kwargs = {}
        if kwarg is None:
            inj_args.append((task, key))
        else:
            inj_kwargs[kwarg] = (task, key)
        LOGGER.debug('inj_args = %s', inj_args)
        LOGGER.debug('inj_kwargs = %s', inj_kwargs)
        return cls(inj_args=inj_args, inj_kwargs=inj_kwargs, wrapped=func,
                   deps_type=deps_type, serialize=serialize)

    def __init__(self, *, inj_args=None, inj_kwargs=None, wrapped,
                 deps_type='hard', serialize=False):
        '''Instantiate a :class:`Use` by providing all  the necessary
        information.

        :param inj_args: an iterable of `(task, key)` pairs specifying which
            part of what task result should be injected as a positional
            argument. See :class:`Use` for the meaning of `key`.
        :type inj_args: tuple((Task, str or None)) or list((Task, str or None))
        :param inj_kwargs: a dictionary associating any kwarg names to a
            `(task, key)` pair. The specified task result will be injected as
            the given kwarg.
        :type inj_kwargs: dict(str, (Task, str or None))
        :param wrapped: the function to wrap.
        :param str deps_type: whether the created task should have a hard or a
            soft dependency towards the injected tasks. Possible values are
            ``'hard'`` and ``'soft'``.
        :param bool serialize: if `True`, the result of this task will be
            written to the output directory.
        '''
        self.inj_args = () if inj_args is None else inj_args
        self.inj_kwargs = {} if inj_kwargs is None else inj_kwargs
        self.wrapped = wrapped
        if isinstance(wrapped, self.__class__):
            func_name = wrapped.func_name
        else:
            func_name = (wrapped.__qualname__ if wrapped.__name__ is None
                         else wrapped.__name__)
            update_wrapper(self, wrapped)
        self.func_name = func_name
        expected_deps_type = ['hard', 'soft']
        if deps_type not in expected_deps_type:
            raise ValueError('deps_type argument to Use must be one of '
                             f'{expected_deps_type}')
        self.deps_type = deps_type
        self.serialize = bool(serialize)

    def get_task(self):
        '''Return the task that executes the decorated function.'''
        from itertools import chain

        if self.deps_type == 'soft':
            soft_deps = set(value[0]
                            for value in chain(self.inj_kwargs.values(),
                                               self.inj_args))
            deps = set()
        else:
            deps = set(value[0] for value in chain(self.inj_kwargs.values(),
                                                   self.inj_args))
            soft_deps = set()

        if deps:
            pytask_name = (','.join(sorted(dep.name for dep in deps))
                           + '.' + self.func_name)
        else:
            pytask_name = self.func_name

        if self._USE_CACHING and pytask_name in self._CACHE:
            LOGGER.debug('returning cached task %s', pytask_name)
            return self._CACHE[pytask_name]

        LOGGER.debug('creating %s task', pytask_name)
        LOGGER.debug('  args: %s', self.inj_args)
        LOGGER.debug('  kwargs: %s', self.inj_kwargs)
        LOGGER.debug('  deps_type: %s', self.deps_type)
        LOGGER.debug('  will depend on %s', deps)
        LOGGER.debug('  will soft-depend on %s', soft_deps)

        def inject_from_env(*, env, config):
            # Prepare the args for the function; we loop over self.inj_args in
            # reverse because stacked decorators are invoked from the inside
            # out. If we do not reverse the list, the outermost decorator will
            # provide the last positional argument.
            args = []
            for task, key in reversed(self.inj_args):
                task_result = from_env(env=env, task_name=task.name, key=key)
                args.append(task_result)

            # prepare the kwargs for the function
            kwargs = {}
            for kwarg, (task, key) in self.inj_kwargs.items():
                task_result = from_env(env=env, task_name=task.name, key=key)
                kwargs[kwarg] = task_result

            LOGGER.debug('injecting args: %s', args)
            LOGGER.debug('injecting kwargs: %s', kwargs)

            result = self.wrapped(*args, **kwargs)
            env_up = {pytask_name: {'result': result}}
            if self.serialize:
                output_dir = Path(config.query('path', 'output-root'),
                                  sanitize_filename(pytask_name))
                ensure(output_dir, is_dir=True)
                env_up[pytask_name]['output_dir'] = str(output_dir)
            return env_up, TaskStatus.DONE

        task = PythonTask(pytask_name, inject_from_env,
                          deps=deps, soft_deps=soft_deps,
                          env_kwarg='env', config_kwarg='config')
        self._CACHE[pytask_name] = task
        return task

    def __call__(self, *args, **kwargs):
        '''Redirect the call to the wrapped function.'''
        return self.wrapped(*args, **kwargs)

    def map(self, func):
        '''Create a new :class:`Use` object that applies `func` to the result
        of the task defined by `self`.'''
        return Use.from_func(func=func, task=self.get_task())


def using(*, key='result', task, kwarg=None):
    '''Make it possible to instantiate :class:`Use` as a decorator.

    See :meth:`Use.__init__` for a description of the parameters.
    '''
    def decorator(wrapped):
        LOGGER.debug('wrapping %s into a Use object', wrapped)
        return Use.from_func(task=task, key=key, kwarg=kwarg, func=wrapped)
    return decorator


class UseRun:
    '''Produce :class:`Use` decorators from a :class:`~.RunTaskFactory`.'''

    @classmethod
    def from_factory(cls, factory):
        '''Create a :class:`UseRun` from a :class:`~.RunTaskFactory`.

        Given a :class:`~.RunTaskFactory`, the :class:`UseRun` class can be
        used to construct :class:`~.RunTask` objects on demand and inject their
        results into the decorated function.

        :param RunTaskFactory factory: a :class:`~.RunTaskFactory` object.
        '''
        return cls(factory, [])

    def __init__(self, factory, posts):
        '''Instantiate a :class:`UseRun` object.

        Given a :class:`~.RunTaskFactory`, the :class:`UseRun` class can be
        used to construct :class:`~.RunTask` objects on demand and inject their
        results into the decorated function.

        The `posts` parameters is a list of functions that will be sequentially
        applied to the result of the generated :class:`~.RunTask`. Each
        function becomes a new :class:`~.PythonTask` object depending on the
        result of the previous one.

        :param RunTaskFactory factory: a :class:`~.RunTaskFactory` object.
        :param list posts: a collection of post-processing functions. Each
            function will be converted to a :class:`~.PythonTask` object.
        '''
        self.factory = factory
        self.posts = posts.copy()

    def __call__(self, kwarg=None, **kwargs):
        task = self.factory.make(**kwargs)
        LOGGER.debug('mapping %s functions on the output of %s',
                     len(self.posts), task.name)
        for post in self.posts:
            use = Use.from_func(task=task, key='result', func=post)
            task = use.get_task()
        return using(kwarg=kwarg, task=task, key='result')

    def copy(self):
        '''Return a copy of this object.'''
        return UseRun(self.factory.copy(), self.posts.copy())

    def map(self, func):
        '''Create a new instance of :class:`UseRun` by extending the list of
        post-processing functions with a new one.

        :param func: a function with one argument to be applied to the result
                     of `self`.
        '''
        copy = self.copy()
        copy.posts.append(func)
        return copy



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module contains code that converts a test result into some kind of
human-readable representation (table, plots, etc.).

.. todo::

    Possible improvement: turn :class:`Representer` into an `ABC`; loop over
    the classes in :mod:`~.eponine` that inherit from `:class:`~.TestResult`
    and add `@abstractmethod` methods in :class:`Representer`. This way, if
    a new :class:`~.TestResult` is added to :mod:`~.eponine`, it will no longer
    be possible to instantiate any of the classes that derive from
    :class:`Representer`, pointing to the fact that the code in this module
    needs to be extended to handle the new class. This is better than silently
    falling back to some default do-nothing implementation, which may lead to
    bugs.


This module uses the `Strategy` design pattern. The :class:`Representation`
class plays the role of `Context`,  :class:`Representer` plays the role of
`Strategy` and the classes derived from :class:`Representer` (such as
:class:`TableRepresenter`, :class:`PlotRepresenter`, etc.) play the role of
`ConcreteStrategy`. See E. Gamma et al., "Design Patterns" (1995),
ISBN 0-201-63361-2, Addison-Weasley, USA.


Currently we have 3 main ``Representer`` classes:

* :class:`Representer`: parent class of all others, containing the default
  :meth:`Representer.__call__` method, calling the class method named
  ``'repr_' + class.name`` of the test;
* :class:`TableRepresenter`: inherited from :class:`Representer`,
  designed as a parent class for user's own representations of tables. Its
  :meth:`TableRepresenter.__call__` method first looks for a method called
  ``'repr_' + class.name``; if it does not exist call the default method from
  the catalogue of table representation accessible in :mod:`.table_repr`;
* :class:`PlotRepresenter`: inherited from :class:`Representer`, designed
  as a parent class for user's own representations of plots. Its
  :meth:`PlotRepresenter.__call__` method first looks for a method called
  ``'repr_' + class.name``; if it does not exist call the default method from
  the catalogue of plot representation accessible in :mod:`.plot_repr`.


An example of use of the ``Representer`` objects can be seen in the
:class:`FullTableRepresenter`. In this table representation, for the Bonferroni
test result, the input test result is first represented in a table, then the
Bonferroni itself is represented in a second table.


Thus the use of the ``Representer`` is foreseen as:

* use the default methods provided in :class:`TableRepresenter` and
  :class:`PlotRepresenter`;
* if customisation is needed, you can easily call the additional method
  available in :mod:`.table_repr` and :mod:`.plot_repr`;
* you can also write your own representation method provided they follow the
  naming convention ``'repr_' + class.name``;
* :mod:`valjean` calls the ``Representer`` classes through the
  :class:`Representation` class.
'''
import logging
from . import table_repr
from . import plot_repr
from .verbosity import Verbosity


LOGGER = logging.getLogger(__name__)


class Representation:
    '''Class for representing test results as templates calling the available
    representers (tables or plots).

    This class corresponds to the `Context` role in the `Strategy` design
    pattern.
    '''

    def __init__(self, representer, verbosity=Verbosity.DEFAULT):
        ''''Initilialisation of the :class:`Representation` class with the
        Representer to use.

        :param Representer: representer to use (table, plots, both, etc)
        :type representer: :class:`Representer`
        '''
        self.representer = representer
        self.verbosity = verbosity

    def __call__(self, result):
        '''Dispatch handling of `result` to the ``__call__`` methods of the
        representer class.'''
        if callable(self.verbosity):
            verbosity = self.verbosity(result)
        else:
            verbosity = self.verbosity
        res = self.representer(result, verbosity)
        if res is None:
            return []
        LOGGER.debug('representing the result of test %s as %s',
                     result.test.name, res)
        return res


class Representer:
    '''Base class for representing test results as templates (in the sense of
    the :mod:`~.templates` module).

    This class corresponds to the `Strategy` role in the `Strategy` design
    pattern. Its subclasses play the role of `ConcreteStrategy`.
    '''

    def __call__(self, result, verbosity=Verbosity.DEFAULT):
        '''Dispatch handling of `result` to the appropriate subclass method,
        based on the name of the class of `result`. This methods essentially
        implements a simplified, run-time version of the Visitor pattern.
        '''
        LOGGER.debug("In Representer.__call__")
        class_name = result.__class__.__name__
        meth_name = 'repr_' + class_name.lower()
        try:
            meth = getattr(self, meth_name)
        except AttributeError:
            LOGGER.debug('no representer for class %s with name %s',
                         class_name, meth_name)
            return None
        return meth(result, verbosity)


class ExternalRepresenter:
    '''This class is the default representation class for external tests, i.e.
    tests defined by the users who already defined the test representation as
    templates.
    '''

    def __call__(self, result, _verbosity):
        LOGGER.debug("In ExternalRepresenter.__call__")
        if result.__class__.__name__ != 'TestResultExternal':
            LOGGER.debug('ExternalRepresenter only deals with '
                         'TestResultExternal')
            return []
        return result.test.templates


class TableRepresenter(Representer):
    '''This class is the default representation class for tables. It contains
    the overridden :meth:`Representer.__call__`.

    Advice: users willing to customize the behaviour of TableRepresenter for
    specific test results should subclass TableRepresenter and define the
    relevant ``repr_*`` methods.
    '''

    def __call__(self, result, verbosity=Verbosity.DEFAULT):
        LOGGER.debug("In TableRepresenter.__call__, %s", verbosity)
        res = super().__call__(result, verbosity)
        if res is None:
            class_name = result.__class__.__name__
            meth_name = 'repr_' + class_name.lower()
            try:
                meth = getattr(table_repr, meth_name)
            except AttributeError:
                LOGGER.info('no table representer %s', meth_name)
                return None
            res = meth(result, verbosity)
        return res


class FullTableRepresenter(TableRepresenter):
    '''Class to define the specific methods for full representation of tables.
    This only involve few cases needing the :meth:`TableRepresenter.__call__`
    method like in Bonferroni and Holm-Bonferroni test results.
    '''

    def repr_testresultbonferroni(self, result, verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestBonferroni` test in two
        tables:

        1. First test result (Student, equal, etc)
        2. Bonferroni test result

        :param  result: a test result.
        :type result: TestResultBonferroni
        :returns: Representation of a :class:`~.TestResultBonferroni` as
            two tables (the first test result and the Bonferroni result).
        :rtype: list(TableTemplate)
        '''
        LOGGER.debug("In FullTableRepresenter.repr_testresultbonferroni")
        if verbosity == Verbosity.SILENT:
            return table_repr.repr_testresultbonferroni(result, verbosity)
        ftest_verb = (Verbosity(verbosity.value-1)
                      if bool(result) else verbosity)
        return (table_repr.repr_testresultbonferroni(result, verbosity)
                + super().__call__(result.first_test_res, ftest_verb))

    def repr_testresultholmbonferroni(self, result,
                                      verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestHolmBonferroni` test in two
        tables:

        1. First test result (Student, equal, etc)
        2. Holm-Bonferroni test result

        :param  result: a test result.
        :type result: TestResultHolmBonferroni
        :returns: Representation of a :class:`~.TestResultHolmBonferroni` as
            two tables (the first test result and the Holm-Bonferroni result).
        :rtype: list(TableTemplate)
        '''
        LOGGER.debug(
            "In FullTableRepresenter.repr_testresultholmbonferroni")
        if verbosity == Verbosity.SILENT:
            return table_repr.repr_testresultholmbonferroni(result, verbosity)
        ftest_verb = (Verbosity(verbosity.value-1)
                      if bool(result) else verbosity)
        return (table_repr.repr_testresultholmbonferroni(result, verbosity)
                + super().__call__(result.first_test_res, ftest_verb))


class PlotRepresenter(Representer):
    '''This class is the default representation class for plots. It contains
    the overridden :meth:`Representer.__call__`.

    Advice: users willing to customize the behaviour of
    :class:`PlotRepresenter` for specific test results should subclass
    :class:`PlotRepresenter` and define the relevant ``repr_*`` methods.
    '''
    def __init__(self, post='default'):
        if callable(post):
            self.post = post
        elif post.lower() == 'none':
            self.post = None
        elif post.lower() == 'default':
            self.post = plot_repr.post_treatment
        else:
            LOGGER.warning('Plot post-treatment must be a callable.')
            self.post = None

    def __call__(self, result, verbosity=Verbosity.DEFAULT):
        LOGGER.debug("In PlotRepresenter.__call__")
        res = super().__call__(result, verbosity)
        if res is None:
            class_name = result.__class__.__name__
            meth_name = 'repr_' + class_name.lower()
            try:
                meth = getattr(plot_repr, meth_name)
            except AttributeError:
                LOGGER.info('no plot representer %s', meth_name)
                return None
            res = meth(result, verbosity)
        if res and self.post is not None:
            res = self.post(res, result)
        return res

    def repr_testresultbonferroni(self, result, verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestBonferroni` test one a plot
        (only the input test for the moment) (Student, equal, etc)

        :param  result: a test result.
        :type result: TestResultBonferroni
        :returns: Representation of a :class:`~.TestResultBonferroni` as a plot
            (the first test result).
        :rtype: list(PlotTemplate)
        '''
        LOGGER.debug("In FullPlotRepresenter.repr_testresultbonferroni")
        return self(result.first_test_res, verbosity)

    def repr_testresultholmbonferroni(self, result,
                                      verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestHolmBonferroni` test as a
        plot (Student, equal, etc)

        :param  result: a test result.
        :type result: TestResultHolmBonferroni
        :returns: Representation of a :class:`~.TestResultHolmBonferroni` as a
            plot (the first test result).
        :rtype: list(PlotTemplate)
        '''
        LOGGER.debug("In PlotRepresenter.repr_testresultholmbonferroni")
        return self(result.first_test_res, verbosity)


class FullPlotRepresenter(PlotRepresenter):
    '''Class to define the specific methods for full representation of plots.
    This only involve few cases needing the :meth:`PlotRepresenter.__call__`
    method like in Bonferroni and Holm-Bonferroni test results.
    '''

    def repr_testresultbonferroni(self, result, verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestBonferroni` test one a plot
        (only the input test for the moment) (Student, equal, etc)

        :param  result: a test result.
        :type result: TestResultBonferroni
        :returns: Representation of a :class:`~.TestResultBonferroni` as a plot
            (the first test result).
        :rtype: list(PlotTemplate)
        '''
        LOGGER.debug("In FullPlotRepresenter.repr_testresultbonferroni")
        return self(result.first_test_res, verbosity)

    def repr_testresultholmbonferroni(self, result,
                                      verbosity=Verbosity.DEFAULT):
        '''Represent the result of a :class:`~.TestHolmBonferroni` test as a
        plot (Student, equal, etc)

        :param  result: a test result.
        :type result: TestResultHolmBonferroni
        :returns: Representation of a :class:`~.TestResultHolmBonferroni` as a
            plot (the first test result).
        :rtype: list(PlotTemplate)
        '''
        LOGGER.debug(
            "In FullPlotRepresenter.repr_testresultholmbonferroni")
        return self(result.first_test_res, verbosity)


class EmptyRepresenter(Representer):
    '''Class that does not generate any templates for any test result.'''


class FullRepresenter(Representer):
    '''This class generates the fullest possible representation for test
    results. If anything can be represented as an template,
    :class:`FullRepresenter` will do it.
    '''

    def __init__(self, post=plot_repr.post_treatment):
        '''Initialisation of :class:`FullRepresenter`.

        Two instance objects are built: a :class:`FullTableRepresenter` and
        a :class:`PlotRepresenter`.
        '''
        LOGGER.debug("In initialisation of FullRepresenter")
        self.table_repr = FullTableRepresenter()
        self.plot_repr = FullPlotRepresenter(post=post)
        self.ext_repr = ExternalRepresenter()

    def __call__(self, result, verbosity=Verbosity.DEFAULT):
        '''Dispatch handling of `result` to all the Representer subclass
        instance attributes of :class:`FullRepresenter`, based on the name of
        the class of `result`.

        If the representer does not exist in :mod:`~.table_repr` or
        :mod:`~.plot_repr` a ``None`` is returned and replaced here by an
        empty list. If none of them exist the global return will be an empty
        list (no ``None`` returned from this step).
        '''
        res = []
        pltres = self.plot_repr(result, verbosity)
        if pltres is not None:
            res.extend(pltres)
        tabres = self.table_repr(result, verbosity)
        if tabres is not None:
            res.extend(tabres)
        extres = self.ext_repr(result, verbosity)
        if extres is not None:
            res.extend(extres)
        return res



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module defines a few tests that report about the success/failure status
of other tests/tasks.

Three different tests are available:

* statistics over all the tasks done, generated with :func:`task_stats`
* statistics over all the tests performed with :func:`test_stats`
* statistics over the tests performed based on labels given at test
  initialization with :func:`test_stats_by_labels`

The two tests over the test results are performed using the
:class:`~.TestResult`, so the statistics only includes the tests that were
actually run.
'''

import logging
from collections import defaultdict
from functools import partial, update_wrapper, total_ordering
from enum import IntEnum

from ...fingerprint import fingerprint
from ...cosette.task import TaskStatus, close_dependency_graph
from ...cosette.use import Use
from ..test import Test, TestResult
from ..eval_test_task import EvalTestTask
from ...eponine.browser import Index


LOGGER = logging.getLogger(__name__)


def stats_worker(test_fn, name, description, tasks, **kwargs):
    '''Function creating the test for all the required tasks (summary tests of
    test tasks or summary tests on all tasks for example).

    :param test_fn: function generating the test to apply to tasks
    :param str name: the name of the stat task
    :param str description: its description
    :param tasks: the list of tasks to be tested.
    :type tasks: list(Task)
    :returns: an :class:`~.EvalTestTask` that evaluates the diagnostic test.
    :rtype: EvalTestTask
    '''
    test_fn.__name__ = name + '.stats'
    test_fn.__qualname__ = test_fn.__qualname__[-10] + name + '.stats'
    inj_args = [(task, None) for task in tasks]
    wrapped = partial(test_fn, name=name, description=description, **kwargs)
    update_wrapper(wrapped, test_fn)
    use_create_test = Use(inj_args=inj_args, wrapped=wrapped, deps_type='soft')
    return EvalTestTask.from_test_task(use_create_test.get_task())


def task_stats(*, name, description='', labels=None, tasks):
    '''Create a :class:`TestStatsTasks` from a list of tasks.

    The :class:`TestStatsTasks` class must be instantiated with the list of
    task results, which are not available to the user when the tests are
    specified in the job file. Therefore, the creation of the
    :class:`TestStatsTasks` must be delayed until the other tasks have finished
    and their results are available in the environment. For this purpose it is
    necessary to wrap the instantiation of :class:`TestStatsTasks` in a
    :class:`~.Use` wrapper, and evaluate the resulting test using a
    :class:`~.EvalTestTask`.

    This function hides this bit of complexity from the user. Assume you have a
    list of tasks that you would like to produce statistics about (we will use
    :class:`~.DelayTask` objects for our example):

    >>> from valjean.cosette.task import DelayTask
    >>> my_tasks = [DelayTask(1), DelayTask(3), DelayTask(0.2)]

    Here is how you make a :class:`TestStatsTasks`:

    >>> stats = task_stats(name='delays', tasks=my_tasks)
    >>> from valjean.gavroche.eval_test_task import EvalTestTask
    >>> isinstance(stats, EvalTestTask)
    True
    >>> print(stats.depends_on)
    {Task('delays.stats')}
    >>> create_stats = next(task for task in stats.depends_on)

    Here `create_stats` is the task that actually creates the
    :class:`TestStatsTasks`. It soft-depends on the tasks in `my_tasks`:

    >>> [task in create_stats.soft_depends_on for task in my_tasks]
    [True, True, True]

    The reason why the dependency is soft is that we want to collect statistics
    about the task outcome in any case, even (especially!) if some of the tasks
    failed.

    :param str name: the name of the task to create.
    :param str description: its description.
    :param tasks: the list of tasks to be tested.
    :type tasks: list(Task)
    :returns: an :class:`~.EvalTestTask` that evaluates the diagnostic test.
    :rtype: EvalTestTask
    '''
    def create_test(*task_results, name, description, labels):
        return [TestStatsTasks(name=name, description=description,
                               labels=labels, task_results=task_results)]

    return stats_worker(create_test, name=name, description=description,
                        labels=labels, tasks=close_dependency_graph(tasks))


@total_ordering
class NameFingerprint:
    '''A small helper class to store a name and an optional fingerprint for the
    referenced item.'''

    def __init__(self, name, fingerprint=None):
        self.name = name
        self.fingerprint = fingerprint

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return repr(self.name)

    def __eq__(self, other):
        return (self.name == other.name
                and self.fingerprint == other.fingerprint)

    def __lt__(self, other):
        if self.name != other.name:
            return self.name < other.name
        return self.fingerprint < other.fingerprint


class TestStatsTasks(Test):
    '''A test that evaluates statistics about the success/failure status of the
    given tasks.
    '''

    def __init__(self, *, name, description='', labels=None, task_results):
        '''Instantiate a :class:`TestStatsTasks`.

        :param str name: the test name.
        :param str description: the test description.
        :param task_results: a list of task results, intended as the contents
            of the environment sections associated with the executed tasks.
            This test notably inspects the ``'status'`` key to see if the task
            succeeded.
        :type task_results: list(dict(str, *stuff*))
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.task_results = task_results

    def evaluate(self):
        '''Evaluate this test and turn it into a :class:`TestResultStatsTasks`.
        '''
        status_dict = defaultdict(list)
        for task_name, task_result in self.task_results:
            name_fing = NameFingerprint(task_name)
            status_dict[task_result['status']].append(name_fing)
        return TestResultStatsTasks(test=self, classify=status_dict)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        for task_name, _task_result in self.task_results:
            yield task_name.encode('utf-8')


class TestResultStatsTasks(TestResult):
    '''The result of the evaluation of a :class:`TestStatsTasks`. The test is
    considered successful if all the observed tasks have successfully completed
    (``TaskStatus.DONE``).
    '''

    def __init__(self, *, test, classify):
        '''Instantiate a :class:`TestResultStatsTasks`.

        :param TestStatsTasks test: the test producing this result.
        :param classify: a dictionary mapping the task status to the list of
            task names with the given status.
        :type classify: dict(TaskStatus, list(str))
        '''
        super().__init__(test=test)
        self.classify = classify

    def __bool__(self):
        '''Returns `True` if all the observed tests have succeeded.'''
        return TaskStatus.DONE in self.classify and len(self.classify) == 1


class TestOutcome(IntEnum):
    '''An enumeration that represents the possible outcomes of a test:

    `SUCCESS`
        represents tests that have been evaluated and have succeeded;

    `FAILURE`
        represents tests that have been evaluated and have failed;

    `MISSING`
        represents tasks that did not generate any ``'result'`` key;

    `NOT_A_TEST`
        represents tasks that did not generate a :class:`~.TestResult` object
        as a result;
    '''
    SUCCESS = 0
    FAILURE = 1
    MISSING = 2
    NOT_A_TEST = 3
    __test__ = False


def test_stats(*, name, description='', labels=None, tasks):
    '''Create a :class:`TestStatsTests` from a list of tests.

    The :class:`TestStatsTests` class must be instantiated with the list of
    test results, which are not available to the user when the tests are
    specified in the job file. Therefore, the creation of the
    :class:`TestStatsTests` must be delayed until the test tasks have finished
    and their results are available in the environment. For this purpose it is
    necessary to wrap the instantiation of :class:`TestStatsTests` in a
    :class:`~.Use` wrapper, and evaluate the resulting test using a
    :class:`~.EvalTestTask`.

    This function hides this bit of complexity from the user. Assume you have a
    list of tasks that evaluate some tests and that you would like to produce
    statistics about the tests results. Let us construct a toy dataset first:

    >>> from collections import OrderedDict
    >>> import numpy as np
    >>> from valjean.eponine.dataset import Dataset
    >>> x = np.linspace(-5., 5., num=100)
    >>> y = x**2
    >>> error = np.zeros_like(y)
    >>> bins = OrderedDict([('x', x)])
    >>> parabola = Dataset(y, error, bins=bins, name='parabola')
    >>> parabola2 = Dataset(y*(1+1e-6), error, bins=bins, name='parabola2')

    Now we write a function that generates dummy tests for the `parabola`
    dataset:

    >>> from valjean.gavroche.test import TestEqual, TestApproxEqual
    >>> def test_generator():
    ...     result = [TestEqual(parabola, parabola2, name='equal?').evaluate(),
    ...               TestApproxEqual(parabola, parabola2,
    ...                               name='approx_equal?').evaluate()]
    ...     return {'test_generator': {'result': result}}, TaskStatus.DONE


    We need to wrap this function in a PythonTask so that it can be executed as
    a part of the dependency graph:

    >>> from valjean.cosette.pythontask import PythonTask
    >>> create_tests_task = PythonTask('test_generator', test_generator)

    Here is how you make a :class:`TestStatsTests` to collect statistics about
    the results of the generated tests:

    >>> stats = test_stats(name='equal', tasks=[create_tests_task])

    >>> from valjean.gavroche.eval_test_task import EvalTestTask
    >>> isinstance(stats, EvalTestTask)
    True

    Here `stats` evaluates the test that gathers the statistics, and it depends
    on a special task that generates the :class:`TestStatsTests` instance:

    >>> print(stats.depends_on)
    {Task('equal.stats')}
    >>> create_stats = next(task for task in stats.depends_on)

    In turn, `create_stats` has a soft dependency on the task that generates
    our test, `create_tests_task`:

    >>> create_tests_task in create_stats.soft_depends_on
    True

    The reason why the dependency is soft is that we want to collect statistics
    about the test outcome in any case, even (especially!) if some of the tests
    failed or threw exceptions.

    Let's run the tests:

    >>> from valjean.config import Config
    >>> config = Config()
    >>> from valjean.cosette.env import Env
    >>> env = Env()
    >>> for task in [create_tests_task, create_stats, stats]:
    ...     env_up, status = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(status)
    TaskStatus.DONE

    The results are stored in a :class:`list` under the key ``'result'``:

    >>> print(len(env[stats.name]['result']))
    1
    >>> stats_res = env[stats.name]['result'][0]
    >>> print("SUCCESS:", stats_res.classify[TestOutcome.SUCCESS])
    SUCCESS: ['approx_equal?']
    >>> print("FAILURE:", stats_res.classify[TestOutcome.FAILURE])
    FAILURE: ['equal?']

    :param str name: the name of the task to create.
    :param str description: its description.
    :param tasks: the list of tasks that generate the tests to observe.
    :type tasks: list(Task)
    :returns: an :class:`~.EvalTestTask` that evaluates the diagnostic test.
    :rtype: EvalTestTask
    '''

    def create_test(*task_results, name, description, labels):
        return [TestStatsTests(name=name, description=description,
                               labels=labels, task_results=task_results)]

    return stats_worker(create_test, name=name, description=description,
                        labels=labels, tasks=tasks)


# hey, pytest!
test_stats.__test__ = False


class TestStatsTests(Test):
    '''A test that evaluates statistics about the success/failure of the given
    tests.
    '''

    def __init__(self, *, name, description='', labels=None, task_results):
        '''Instantiate a :class:`TestStatsTests` from a collection of task
        results. The tasks are expected to generate :class:`~.TestResult`
        objects, which must appear in the ``'result'`` key of the task result.
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.task_results = task_results

    def evaluate(self):
        '''Evaluate this test and turn it into a :class:`TestResultStatsTests`.
        '''
        status_dict = defaultdict(list)
        for task_name, task_result in self.task_results:
            if 'result' not in task_result:
                name_fing = NameFingerprint(task_name)
                status_dict[TestOutcome.MISSING].append(name_fing)
                continue
            test_results = task_result['result']
            for test_result in test_results:
                if not isinstance(test_result, TestResult):
                    name_fing = NameFingerprint(task_name)
                    status_dict[TestOutcome.NOT_A_TEST].append(name_fing)
                if test_result:
                    test_lst = status_dict[TestOutcome.SUCCESS]
                else:
                    test_lst = status_dict[TestOutcome.FAILURE]
                name_fing = NameFingerprint(test_result.test.name,
                                            fingerprint(test_result.test))
                test_lst.append(name_fing)
        return TestResultStatsTests(test=self, classify=status_dict)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        for task_name, _task_result in self.task_results:
            yield task_name.encode('utf-8')


class TestResultStatsTests(TestResultStatsTasks):
    '''The result of the evaluation of a :class:`TestStatsTests`. The test is
    considered successful if all the observed tests have been successfully
    evaluated and have succeeded.
    '''
    def __bool__(self):
        return TestOutcome.SUCCESS in self.classify and len(self.classify) == 1


def test_stats_by_labels(*, name, description='', labels=None,
                         tasks, by_labels):
    '''Create a :class:`TestStatsTestsByLabels` from a list of tests.

    See :func:`test_stats` for the generalities about this function.

    Compared to :func:`test_stats` it takes one argument more: ``'by_labels'``
    to classify then build statistics based on these labels. **The order of the
    labels matters**, as they are successively selected.

    Let's define three menus:

    >>> menu1 = {'food': 'egg + spam', 'drink': 'beer'}
    >>> menu2 = {'food': 'egg + bacon', 'drink': 'beer'}
    >>> menu3 = {'food': 'lobster thermidor', 'drink': 'brandy'}

    These three menus are ordered by pairs. Statistics on meals are
    kept in the restaurant, using :class:`~.TestMetadata`. The goal of the
    tests is to know if both persons of a pair order the same menu and when
    they do it.

    .. code::

        orders = [TestMetadata(
            {'Graham': menu1, 'Terry': menu1}, name='gt_wday_lunch',
            labels={'day': 'Wednesday', 'meal': 'lunch'}),
                  TestMetadata(
            {'Michael': menu1, 'Eric': menu2}, name='me_wday_dinner',
            labels={'day': 'Wednesday', 'meal': 'dinner'}),
                  TestMetadata(
            {'John': menu2, 'Terry': menu2}, name='jt_wday',
            labels={'day': 'Wednesday'}),
                  TestMetadata(
            {'Terry': menu3, 'John': menu3}, name='Xmasday',
            labels={'day': "Christmas Eve"})]

    The restaurant owner uses :func:`test_stats_by_labels` to build statistics
    on his menus and the habits of his consumers.

    For example, the menus filtered on ``day`` will give:

    .. code::

        =============  ============  =============
              day       % success      % failure
        =============  ============  =============
        Christmas Eve      1/1            0/1
            Wednesday      2/3            1/3
        =============  ============  =============

    These results means, considering the tests requested, both consumers have
    the same meal on Christmas Eve. On Wednesday, one pair of customers out of
    three did not order the same menu.

    The same kind of statistics can be done based on the meal:

    .. code::

        ==========  ============  =============
           meal       % success     % failure
        ==========  ============  =============
          dinner         0/1            1/1
           lunch         1/1            0/1
        ==========  ============  =============

    In that case two tests were not taken into account as they did not have any
    ``'meal'`` label.

    It is also possible to make selections on multiple labels. In that case the
    order matters: the classification is performed following the order of the
    labels requested. For example, ``'meal'`` then ``'day'`` :

    .. code::

        ==========  =========  ============  =============
           meal        day       % success     % failure
        ==========  =========  ============  =============
          dinner    Wednesday       0/1            1/1
           lunch    Wednesday       1/1            0/1
        ==========  =========  ============  =============

    * Only two tests are filtered due to the meal selection
    * Requesting ``'day'`` then ``'meal'`` would only inverse the two first
      columns in that case and emit a **warning**: a preselection on ``'day'``
      is done and in Christmas Eve case the ``'meal'`` label is not provided,
      the selection cannot be done. In the Wednesday one, no problem ``'meal'``
      appears at least in one case (two in our cases).

    Finally if the request involves a label that does not exist in any test an
    exception will be raised, mentioning the failing label.

    :param str name: the name of the task to create.
    :param str description: its description.
    :param tasks: the list of tasks that generate the tests to observe.
    :type tasks: list(Task)
    :param tuple(str) by_labels: labels from the tests on which the
        classification will be based
    :returns: an :class:`~.EvalTestTask` that evaluates the diagnostic test.
    :rtype: EvalTestTask
    '''
    def create_test(*task_results, name, description, labels, by_labels=None):
        return [TestStatsTestsByLabels(
            name=name, description=description, labels=labels,
            task_results=task_results, by_labels=by_labels)]

    return stats_worker(create_test, name=name, description=description,
                        labels=labels, tasks=tasks, by_labels=by_labels)


# hey, pytest!
test_stats_by_labels.__test__ = False


class TestStatsTestsByLabelsException(Exception):
    '''Exception raised during the diagnostic test on :class:`~.TestResult`
    when a classification by labels is required.'''
    # And for pytest...
    __test__ = False


class TestStatsTestsByLabels(Test):
    '''A test that evaluates statistics about the success/failure of the given
    tests using their labels to classify them.

    Usually more than one test is performed for each tested case. This test
    summarize tests done on a given category defined by the user in the usual
    tests (:class:`~.TestStudent`, :class:`~.TestMetadata`, etc.).

    During the evaluation a list of dictionaries of labels is built for each
    test. These labels are given by the user at the initialization of the test.
    Each dictionary also contains the name of the test (name of the task) and
    its result (sucess or failure). From this list of dictionaries an
    :class:`~.Index` is built.

    The result of the evaluation is given a a list of dictionaries containing
    the strings corresponding to the chosen labels under the ``'labels'`` key
    and the number of results OK, KO and total.
    '''
    def __init__(self, *, name, description='', labels=None, task_results,
                 by_labels):
        '''Instantiate a :class:`TestStatsTestsByLabels` from a collection of
        task results. The tasks are expected to generate :class:`~.TestResult`
        objects, which must appear in the ``'result'`` key of the task result.

        :param str name: the test name
        :param str description: the test description
        :param task_result: a list of task results, each task normally contains
            a :class:`~.TestResult`, used in this test.
        :param tuple by_labels: ordered labels to sort the test results. These
            labels are the test labels.
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.task_results = task_results
        self.by_labels = by_labels

    def _build_labels_lod(self):
        '''Build the labels list of dictionaries that allows the creation of
        the index.

        Two additional labels are added to the labels list: ``'_test_name'``
        and ``'_result'`` that should not be used as test's labels. The first
        one could be needed some time and is expected to be unique, the second
        one is filled with :class:`TestOutcome`.
        '''
        labels_lod = []
        for _task_name, task_result in self.task_results:
            if 'result' not in task_result:
                continue
            test_results = task_result['result']
            for test_result in test_results:
                ldic = test_result.test.labels.copy()
                if '_test_name' in ldic:
                    LOGGER.warning(
                        "'_test_name' is a label in some test, will be "
                        "replaced in TestStatsTestsByLabels")
                ldic['_test_name'] = test_result.test.name
                if '_result' in ldic:
                    LOGGER.warning(
                        "'_result' is a label in some test, will be replaced"
                        "in TestStatsTestsByLabels")
                if test_result:
                    ldic['_result'] = TestOutcome.SUCCESS
                else:
                    ldic['_result'] = TestOutcome.FAILURE
                labels_lod.append(ldic)

        return labels_lod

    @staticmethod
    def _build_index(labels_lod):
        index = Index()
        for itres, tres in enumerate(labels_lod):
            for lab in tres:
                index[lab][tres[lab]].add(itres)
        return index

    def _rloop_over_labels(self, index, labels, rok, rko, plab=()):
        '''Recursive method to select labels and calculate the relative
        statistics.

        If ``labels`` contains only one element, the method build a dictionary
        that summarizes the OK, KO and total number of tests for the considered
        label. The method returns a list containing the dictionary as its only
        element.

        The previous labels are obtained from the ``plab`` argument.

        If ``labels`` contains multiple labels, the first label is considered
        first. For each possible value of the label, the method constructs a
        sub-index by filtering the index on the given label and recursively
        calls itself on the sub-index. The ``plab`` arguement contains the list
        of all the labels that have been considered so far. The lists of
        dictionaries that result from the recursive calls are concatenated into
        a single list and returned.

        :param Index index: index used to calculate the statistics
        :param tuple(str) labels: ordered labels to sort the test results
        :param set rok: successful test results
        :param set rko: failed test results
        :param tuple plab: previous labels
        :rtype: list(dict)
        '''
        label = labels[0]
        if label not in index:
            LOGGER.warning('%s not found in some test labels', label)
            return []
        if len(labels) > 1:
            lres = []
            for lab, labset in index.get(label, {}).items():
                subind = index.keep_only(labset)
                lres.extend(self._rloop_over_labels(subind, labels[1:],
                                                    rok, rko,
                                                    plab=plab+(lab,)))
            return lres
        lres = []
        for lab, labset in index.get(label, {}).items():
            lres.append({'labels': plab + (lab,),
                         'OK': len(labset & rok),
                         'KO': len(labset & rko),
                         'total': len(labset)})
        return lres

    def _stats_for_labels(self, index):
        '''Check the presence of all required labels in the index, build the
        success and failures sets and return the dictionary of results.

        :raises: TestStatsTestsByLabelsExeption
        :returns: list of dictionaries of labels and states
        :rtype: list(dict)
        '''
        if not set(self.by_labels) <= set(index):
            raise TestStatsTestsByLabelsException(
                f'TestStatsTestsByLabels: {self.by_labels} not found in test '
                'labels')
        rok = index['_result'][TestOutcome.SUCCESS]
        rko = index['_result'][TestOutcome.FAILURE]
        res = self._rloop_over_labels(index, self.by_labels, rok, rko)
        return sorted(res, key=lambda x: x['labels'])

    def evaluate(self):
        '''Evaluate this test and turn it into a
        :class:`TestResultStatsTestsByLabels`.
        '''
        labels_lod = self._build_labels_lod()
        index = self._build_index(labels_lod)
        sfl = self._stats_for_labels(index)
        return TestResultStatsTestsByLabels(test=self, classify=sfl,
                                            n_labels=len(labels_lod))

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        for task_name, _task_result in self.task_results:
            yield task_name.encode('utf-8')
        for label in self.by_labels:
            yield label.encode('utf-8')


class TestResultStatsTestsByLabels(TestResultStatsTasks):
    '''The result of the evaluation of a :class:`TestStatsTestsByLabels`. The
    test is considered successful if all the observed tests have been
    successfully evaluated and have succeeded.

    An oracle is available for each individual test (usually what is required
    here).

    ``self.classify`` is here a list of dictionaries with the following keys:
    ``['labels', 'OK', 'KO', 'total']``.
    '''

    def __init__(self, *, test, classify, n_labels):
        super().__init__(test=test, classify=classify)
        self.n_labels = n_labels

    def __bool__(self):
        '''Test is successful if all tests are.'''
        return all(self.oracles())

    def oracles(self):
        '''Test if each test is successful.

        :rtype: list(bool)
        '''
        return [t['OK'] == t['total'] for t in self.classify]

    def nb_missing_labels(self):
        '''Return the number of tests where at least one of the labels required
        were missing.

        :rtype: int
        '''
        return self.n_labels - sum(s['total'] for s in self.classify)


def classification_counts(classify, status_first):
    '''Count the occurrences of different statuses in the `classify`
    dictionary.

    :param dict classify: a dictionary associating *things* to statuses. The
        statuses must have the same type as `status_first`
    :param status_first: the status that is considered as success. This must be
        an enum class
    :returns: a pair of lists of equal length. The first element of the pair is
        the list of statuses appearing in `classify` (`status_first` is
        guaranteed to come first in this list); the second element is the
        number of times the corresponding status appears in `classify`.
    '''
    statuses = [status_first]
    statuses.extend(status for status in status_first.__class__
                    if status != status_first)
    counts = [len(classify[status]) for status in statuses]
    statuses = [status for status, count in zip(statuses, counts)
                if count != 0]
    counts = [count for count in counts if count != 0]
    return statuses, counts



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
'''This module provides classes to schedule the execution of several tasks,
possibly dependent on each other.

Example usage:

   >>> from valjean.cosette.task import DelayTask
   >>> spam = DelayTask('spam', 0.1)
   >>> eggs = DelayTask('eggs', 0.2)
   >>> bacon = DelayTask('bacon', 0.2)

   >>> from valjean.cosette.depgraph import DepGraph
   >>> g = DepGraph.from_dependency_dictionary({
   ...         spam: [],
   ...         bacon: [spam],
   ...         eggs: []
   ...     })

   >>> from valjean.cosette.scheduler import Scheduler
   >>> s = Scheduler(hard_graph=g)
   >>> env = s.schedule()  # executes the tasks in the correct order
'''

import logging
from .depgraph import DepGraph
from .backends.queue import QueueScheduling
from .env import Env
from ..chrono import Chrono
from ..config import Config


LOGGER = logging.getLogger(__name__)


class SchedulerError(Exception):
    '''An error that may be raised by the :class:`Scheduler` class.'''


class Scheduler:
    '''Schedule a number of tasks.

    The Scheduler class has the responsibility of scheduling and executing a
    number of tasks.  Here `hard_graph` is a :class:`~.depgraph.DepGraph`
    describing the *hard* dependencies among the tasks to be executed (see
    :mod:`~.task` for details about *hard* vs. *soft* dependencies);
    `soft_graph` is a :class:`~.depgraph.DepGraph` describing the *soft*
    dependencies alone; finally, `backend` should be an instance of a
    `*Scheduling` class such as :class:`.QueueScheduling`, or at any rate a
    class that exhibits an :meth:`~.execute_tasks()` method with the correct
    signature (see :meth:`.QueueScheduling.execute_tasks()`). If `backend` is
    `None`, the default backend will be used.

    :param DepGraph hard_graph: The dependency graph for hard dependencies.
    :param soft_graph: The dependency graph for soft dependencies.
    :type soft_graph: DepGraph or None
    :param backend: The scheduling backend.
    :type backend: None or QueueScheduling
    :raises ValueError: if `depgraph` is not an instance of
                        :class:`~.DepGraph`.
    :raises SchedulerError: if the tasks do not have any ``do()`` method.
    '''

    def __init__(self, *, hard_graph, soft_graph=None, backend=None):
        '''Initialize the scheduler with a graph.'''

        if not isinstance(hard_graph, DepGraph):
            raise ValueError('Scheduler must be initialised with a '
                             'DepGraph (`hard_graph`)')

        if not isinstance(soft_graph, (DepGraph, type(None))):
            raise ValueError('Scheduler must be initialised with a '
                             'DepGraph or None (`soft_graph`)')

        soft_graph = DepGraph() if soft_graph is None else soft_graph.copy()
        with Chrono() as chrono:
            self.hard_graph = hard_graph.copy()
        LOGGER.info('hard graph copied in %s seconds', chrono)
        with chrono:
            self.hard_graph.flatten()
        LOGGER.info('hard graph flattened in %s seconds', chrono)
        with chrono:
            self.full_graph = hard_graph + soft_graph
        LOGGER.info('full graph computed in %s seconds', chrono)
        with chrono:
            self.full_graph.flatten()
        LOGGER.info('full graph flattened in %s seconds', chrono)

        # make sure that all nodes appear in the hard graph
        for node in self.full_graph.nodes():
            self.hard_graph.add_node(node)

        # check that all the nodes of the graph can be executed (i.e. they
        # should have a do() method)
        for node in self.full_graph.nodes():
            if not hasattr(node, 'do') or not hasattr(node.do, '__call__'):
                raise SchedulerError('Cannot schedule tasks for execution: '
                                     f'node {node} does not have a do() '
                                     'method')

        if backend is None:
            self.backend = QueueScheduling()
        else:
            self.backend = backend

    def schedule(self, *, config=None, env=None):
        '''Schedule the tasks!

        :param Env env: An initial environment for the scheduled tasks. This
            allows completed tasks to inform other, dependent tasks of e.g.
            the location of interesting files.
        :param Config config: the configuration object.
        :returns: The modified environment.
        '''
        if config is None:
            config = Config()

        if env is None:
            env = Env()
        LOGGER.info('scheduling tasks')
        LOGGER.debug('for full graph %s', self.full_graph)
        LOGGER.debug('for soft graph %s', self.hard_graph)
        LOGGER.debug('with env %s', env)
        LOGGER.debug('with config %s', config)
        self.backend.execute_tasks(full_graph=self.full_graph,
                                   hard_graph=self.hard_graph,
                                   env=env, config=config)
        return env



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module contains a simple class that models a test report. A report
consists of a number of sections, and each section contains:
* a title;
* some optional introductory text;
* optionally, other test reports and some test results.

This module also contains a task that creates a report from a list of
test-evaluation tasks.
'''

from ..cosette.task import TaskStatus
from ..cosette.pythontask import PythonTask
from ..gavroche.test import TestResult


class TestReport:
    '''This class models a test report. It is essentially a rose tree (i.e. a
    tree with an arbitrary number of children at each node) with nodes
    annotated with a title and some introductory text. :class:`~.TestResult`
    objects play the role of tree leaves.
    '''

    # tell pytest that this class and derived classes should NOT be collected
    # as tests
    __test__ = False

    def __init__(self, *, title, text='', content=None):
        '''Create a report.

        :param str title: the report title.
        :param str text: some introductory text.
        :param content: a list of :class:`TestReport` or :class:`~.TestResult`
                        objects.
        :type content: list(TestReport or TestResult)
        '''
        self.title = title
        self.text = text
        self.content = [] if content is None else content
        for sec in self.content:
            if not isinstance(sec, (TestReport, TestResult)):
                raise TypeError('expected a list of TestReport or TestResult '
                                'objects, but one of the items is a '
                                f'{type(sec)} instead')


class TestReportTask(PythonTask):
    '''Task class that creates a :class:`TestReport` from a function that
    classifies test results into report sections.
    '''

    # tell pytest that this class and derived classes should NOT be collected
    # as tests
    __test__ = False

    def __init__(self, name, *, make_report, eval_tasks, kwargs=None):
        '''Instantiate a :class:`TestReportTask` from a list of
        :class:`~.EvalTestTask` objects. The `make_report` argument is expected
        to be a function taking one argument, which is a dictionary associating
        the names of the tasks in `eval_tasks` to the results of their
        execution (in the sense of the ``'result'`` key of the associated
        environment section). For tasks that evaluate tests
        (:class:`~.EvalTestTask`), the result typically consists of a list of
        :class:`~.TestResult` objects. The `make_report` function must return a
        :class:`TestReport` object.

        :param str name: the name of this task.
        :param make_report: a function taking a single argument and returning a
            :class:`TestReport`.
        :param eval_tasks: a list of tasks, typically :class:`~.EvalTestTask`.
        :type eval_tasks: list(Task)
        :param kwargs: a dictionary of kwargs that will be passed to the
            `make_report` function.
        :type kwargs: dict or None
        '''
        kwargs = {} if kwargs is None else kwargs.copy()

        def report(*, env):
            result_dict = {}
            for test_task in eval_tasks:
                tname = test_task.name
                if (tname not in env
                        or env[tname]['status'] != TaskStatus.DONE
                        or 'result' not in env[tname]):
                    continue  # skip failed or missing tasks
                result_dict[tname] = env[tname]['result']
            report = make_report(result_dict, **kwargs)
            env_up = {self.name: {'result': report}}
            return env_up, TaskStatus.DONE

        super().__init__(name, report, soft_deps=eval_tasks,
                         env_kwarg='env')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
'''This module implements a task class called :class:`PythonTask`, which is
able to execute arbitrary Python code (in the form of a function call) when
requested to do so. This is extra useful because functions wrapped in
:class:`PythonTask` can receive the scheduling environment as an additional
parameter, which makes it possible to write code that depends on the results of
previous tasks.

:class:`PythonTask` objects
---------------------------

Creating a :class:`PythonTask` is very simple. Let us define a function that
returns a constant value:

    >>> def func():
    ...   return 42

We can wrap the function in a :class:`PythonTask` as follows:

    >>> task = PythonTask('answer', func)  # 'answer' is the task name

We can then execute the task by passing an empty environment (dictionary) and
an empty :class:`~.Config` to the :meth:`PythonTask.do` method:

    >>> task.do(env={}, config=None)
    42

Note that :meth:`PythonTask.do` simply returns the result of the wrapped
function. For the sake of illustration, our function returns an integer;
however, if you want to use :class:`PythonTask` in a :class:`~.DepGraph`, your
wrapped function will need to return an `(env_up, status)` pair, like the other
tasks.

.. note::

    Exceptions raised by the wrapped function are not caught by the task (they
    should be caught by the scheduler, though).

Calling a function without any arguments is not very restrictive per se.
Say you want to call a function of two arguments:

    >>> def add(x, y):
    ...   return x + y

If you have your arguments lying around at the time you construct your task,
then you may do something like

    >>> x, y = 5, 3
    >>> task_add = PythonTask('add', lambda: add(x, y))
    >>> task_add.do(env={}, config=None)
    8

Essentially, you construct a trampoline: the lambda takes no arguments, but it
captures `x` and `y` from the surrounding scope and passes them to the
:func:`!add` function. This works, but due to the way Python handles captured
variables, it may bring a few surprises:

    >>> x, y = 5, 3
    >>> task_add = PythonTask('add', lambda: add(x, y))
    >>> x, y = 1, 2
    >>> task_add.do(env={}, config=None)  # this still returns 8, right?
    3
    >>> # WAT

So, unless you know what you are doing, it is better to avoid this surprising
behaviour and use the `args` argument to :class:`PythonTask`:

    >>> x, y = 5, 3
    >>> task_add = PythonTask('add', add, args=(x, y))
    >>> task_add.do(env={}, config=None)
    8

There is also a `kwargs` argument that can be used to pass keyword arguments:

    >>> task_add = PythonTask('add', add, kwargs={'x': x, 'y': y})
    >>> task_add.do(env={}, config=None)
    8


Passing arguments via the environment
-------------------------------------

Sometimes your function requires some arguments, but the arguments themselves
are not available (e.g. they haven't been computed yet) by the time you create
your :class:`PythonTask`. For this purpose, :class:`PythonTask` provides an
additional feature that allows the called function to query the task
environment and retrieve any additional information from there.

The mechanism is simple: the environment is passed to the wrapped function as a
keyword argument. The keyword can be specified by the user using the
`env_kwarg` argument to the :class:`PythonTask` constructor.

As a simple example, consider the following function:

    >>> def goodnight(*, some_dict):
    ...   number = some_dict['number']
    ...   return 'Goodnight, ' + ('ding'*number)

You can wrap it in a :class:`PythonTask` as follows:

    >>> task_gnight = PythonTask('goodnight', goodnight, env_kwarg='some_dict')

and this is how it works:

    >>> env = {'number': 8}
    >>> task_gnight.do(env, config=None)
    'Goodnight, dingdingdingdingdingdingdingding'

Passing arguments via the environment: a more complex example
-------------------------------------------------------------

As an illustration of a more complex scenario, we will now implement a set of
:class:`PythonTask` objects to calculate the Pascal's triangle. In plain
Python, the code to print all the rows up to the n-th would look something like
this:

    >>> import numpy as np
    >>> def pascal(n):
    ...   res = np.zeros((n, n), dtype=int)
    ...   res[:, 0] = 1
    ...   res[0, :] = 1
    ...   for i in range(2, n):
    ...     for j in range(1, i):
    ...       res[i-j, j] = res[i-j-1, j] + res[i-j, j-1]
    ...   return res
    >>> direct_pascal = pascal(8)
    >>> print(direct_pascal)
    [[ 1  1  1  1  1  1  1  1]
     [ 1  2  3  4  5  6  7  0]
     [ 1  3  6 10 15 21  0  0]
     [ 1  4 10 20 35  0  0  0]
     [ 1  5 15 35  0  0  0  0]
     [ 1  6 21  0  0  0  0  0]
     [ 1  7  0  0  0  0  0  0]
     [ 1  0  0  0  0  0  0  0]]

The logic is that each matrix element (except for those in the first
row/column) is the sum of the element above and the element on the left.

In order to compute Pascal's triangle using :class:`PythonTask` objects, we
first need to decide on a strategy. We decide to use a :class:`PythonTask` per
matrix element. We also have to choose a strategy for naming the tasks, because
the content of the environment is indexed by the task name. So we decide to
call ``'(i, j)'`` the task that computes element `(i, j)`.

Armed with these conventions, we can write the function that computes element
`(i, j)`:

    >>> from valjean.cosette.task import TaskStatus
    >>> def compute(name, i, j, *, env):
    ...   if i == 0 or j == 0:
    ...     env_up = {name: {'result': 1}}
    ...     return env_up, TaskStatus.DONE
    ...   left = str((i-1, j))
    ...   above = str((i, j-1))
    ...   left_result = env[left]['result']
    ...   above_result = env[above]['result']
    ...   result = left_result + above_result
    ...   env_up = {name: {'result': result}}
    ...   return env_up, TaskStatus.DONE

Note that we have to return an environment update and a status.  Now we
construct the tasks and assemble them into a dependency dictionary:

    >>> deps = {}
    >>> name_to_task = {}
    >>> n = 8
    >>> for k in range(n):
    ...   # k is the index of the row in the triange
    ...   # i and j index the matrix element, so k = i + j
    ...   for i in range(k+1):
    ...     j = k - i
    ...     task_name = str((i, j))
    ...     task = PythonTask(task_name, compute, args=(task_name, i, j),
    ...                       env_kwarg='env')
    ...     name_to_task[task_name] = task
    ...     deps[task] = set()
    ...     if i > 0:
    ...       index_left = str((i-1, j))
    ...       deps[task].add(name_to_task[index_left])
    ...     if j > 0:
    ...       index_above = str((i, j-1))
    ...       deps[task].add(name_to_task[index_above])

We can then import :class:`~.DepGraph` and :class:`~.Scheduler` and execute the
dependency graph:

    >>> from valjean.cosette.depgraph import DepGraph
    >>> graph = DepGraph.from_dependency_dictionary(deps)
    >>> from valjean.cosette.scheduler import Scheduler
    >>> scheduler = Scheduler(hard_graph=graph)
    >>> final_env = scheduler.schedule()

And now we can extract the results from the final environment:

    >>> pythontask_pascal = np.zeros_like(direct_pascal)
    >>> for k in range(n):
    ...   for i in range(k+1):
    ...     j = k - i
    ...     task_name = str((i, j))
    ...     pythontask_pascal[i, j] = final_env[task_name]['result']
    >>> print(pythontask_pascal)
    [[ 1  1  1  1  1  1  1  1]
     [ 1  2  3  4  5  6  7  0]
     [ 1  3  6 10 15 21  0  0]
     [ 1  4 10 20 35  0  0  0]
     [ 1  5 15 35  0  0  0  0]
     [ 1  6 21  0  0  0  0  0]
     [ 1  7  0  0  0  0  0  0]
     [ 1  0  0  0  0  0  0  0]]
    >>> np.all(pythontask_pascal == direct_pascal)
    True

Passing arguments via the configuration
---------------------------------------

The function wrapped in a :class:`PythonTask` can also inspect the global
`valjean` configuration object. This may be useful to retrieve global settings
for paths, for instance. Like the environment, you can specify that the
configuration should be passed to the wrapped function as a keyword argument.
The keyword is specified by the `config_kwarg` parameter to the
:class:`PythonTask` constructor. For example:

    >>> from valjean.config import Config
    >>> def print_output_dir(*, config):
    ...     print(config.query('path', 'output-root'))
    >>> task = PythonTask('output-dir', print_output_dir,
    ...                   config_kwarg='config')
    >>> config = Config()
    >>> task.do(env={}, config=config)
    /.../output


Module API
----------
'''

import logging
import copy
from .task import Task, TaskStatus


LOGGER = logging.getLogger(__name__)


class TaskException(Exception):
    '''An exception that can be raised by any functions wrapped in
    :class:`PythonTask`. It causes the task to fail. A reason can be specified
    in the constructor.'''

    def __init__(self, reason='unknown'):
        '''Construct a :class:`TaskException`.

        :param str reason: why this exception was raised.'''
        super().__init__()
        self.because = reason


class PythonTask(Task):
    '''Task that executes specified Python code.'''
    def __init__(self, name, func, *, args=None, kwargs=None,
                 env_kwarg=None, config_kwarg=None, deps=None, soft_deps=None):
        # pylint: disable=too-many-arguments
        '''Initialize the task with a function, a tuple of arguments and a
        dictionary of kwargs.

        :param str name: The name of the task.
        :param func: A function to be executed.
        :param tuple args: A tuple of positional arguments to `func`, or `None`
                           if none are required.
        :param dict kwargs: A dictionary of keyword arguments to `func`, or
                            `None` if none are required.
        :param str env_kwarg: The name of the keyword argument that will be
                              used to pass the environment to the function, or
                              `None` if the environment should not be passed.
        :param str config_kwarg: The name of the keyword argument that will be
                                 used to pass the config to the function, or
                                 `None` if the config should not be passed.
        :param deps: If this task depends on other tasks (and valjean cannot
                     automatically discover this), pass them (as a list) to the
                     `deps` parameter.
        :type deps: None or collection of :class:`~.Task` objects.
        :param soft_deps: If this task has a soft dependency on other tasks
                          (and valjean cannot automatically discover this),
                          pass them (as a list) to the `soft_deps` parameter.
        :type soft_deps: None or collection of :class:`~.Task` objects.
        '''
        super().__init__(name=name, deps=deps, soft_deps=soft_deps)
        self.func = func
        self.args = copy.deepcopy(args) if args is not None else ()
        self.kwargs = copy.deepcopy(kwargs) if kwargs is not None else {}
        self.env_kwarg = env_kwarg
        self.config_kwarg = config_kwarg

    def do(self, env, config):
        '''Execute the function.

        :param env: The environment. It will be passed to the executed function
                    as the `env_kwarg` keyword, if specified.
        :param config: The config. It will be passed to the executed function
                       as the `config_kwarg` keyword, if specified.
        '''
        from types import MappingProxyType
        LOGGER.debug('PythonTask %r runs', self.name)
        if self.env_kwarg is not None:
            # wrap the environment in MappingProxyType, so that self.func
            # cannot modify it
            self.kwargs[self.env_kwarg] = MappingProxyType(env)
        if self.config_kwarg is not None:
            self.kwargs[self.config_kwarg] = config
        try:
            # here result is actually an (env_up, status) tuple
            result = self.func(*self.args, **self.kwargs)
        except TaskException as task_exc:
            env_up = {self.name: {'why': task_exc.because}}
            return env_up, TaskStatus.FAILED
        return result
